# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Module with utility and helper functions."""

import asyncio
import base64
import contextlib
import datetime
import json
import logging
import os
import subprocess
import sys
from functools import wraps
from xml.dom.minidom import Document as xml_doc

import yaml

from mrack.errors import ConfigError, ProvisioningError

logger = logging.getLogger(__name__)


def value_to_bool(value):
    """Convert Python variable value to boolean."""
    _false_set = {"no", "false", "f", "n", "0"}
    _true_set = {"yes", "true", "t", "y", "1"}
    if isinstance(value, str):
        if value.lower() in _true_set:
            return True

        if value.lower() in _false_set:
            return False

        raise ValueError(
            f"Invalid input string for boolean conversion '{value}',"
            "supported values are:\n"
            f"False:\t{_true_set}\n"
            f"True:\t{_false_set}"
        )

    return bool(value)


def get_config_value(config_dict, key, default=None):
    """Get configuration value or default value from dictionary.

    Usable for dictionaries which can also contain attribute 'default'

    Order of preference:
    * attribute value
    * value of 'default' attribute
    * value of provided default param
    """
    val = config_dict.get(key)
    if val is None:
        val = config_dict.get("default", default)
    return val


def get_value_or_dict_value(config_dict, attr, dict_name, key):
    """
    Look for a value in configuration dictionary.

    Order of preference:
    * dictionary 'dict_name' with key 'key'
    * attribute: 'attr'
    * None
    """
    if config_dict is None:
        return None

    value = None
    if dict_name and key:
        value_dict = config_dict.get(dict_name)
        if value_dict:
            value = get_config_value(value_dict, key)
            if value:
                return value
        # avoid to return the whole value_dict if dict name is the same as attr
        if attr == dict_name:
            return None

    value = config_dict.get(attr)
    return value


def find_value_in_config_hierarchy(
    provisioning_config,
    provider_key,
    host,
    meta_host,
    attr,
    dict_name,
    key,
    default=None,
):
    """
    Look for first value in configuration hierarchy.

    Order of preference:
    * host value (attribute: attr)
    * host metadata value (attribute: 'attr')
    * provider config value (dictionary 'dict_name' with key 'key')
    * provider default value (attribute: 'attr')
    * global config value (dictionary 'dict_name' with key 'key')
    * global default value (attribute: 'attr')
    * value of provided default param
    """
    value = None
    if host is not None:
        if hasattr(host, attr):
            value = getattr(host, attr)

    if value is None and isinstance(meta_host, dict):
        value = meta_host.get(attr)

    if value is None and provider_key is not None:
        provider_config = provisioning_config.get(provider_key)
        value = get_value_or_dict_value(provider_config, attr, dict_name, key)

    if value is None:
        value = get_value_or_dict_value(provisioning_config, attr, dict_name, key)
    if value is None:
        value = default

    return value


def validate_dict_attrs(dct, attr_list, dct_name):
    """Validate that dictionary contains all attributes.

    Based on provided attribute list.
    """
    missing = [a for a in attr_list if a not in dct]
    if missing:
        error = f"""
        '{dct_name} definition:'
        {dct}
        'Is missing required attributes: {missing}'
        """
        raise ConfigError(error)


def add_dict_to_node(node, input_dict):
    """Convert dict object to XML elements."""
    # Recursively create xml from dict
    if isinstance(input_dict, dict):
        for key, value in input_dict.items():
            if key.startswith("_"):
                node.setAttribute(key[1:], str(value))
            else:
                node.appendChild(add_dict_to_node(xml_doc().createElement(key), value))

    return node


def json_convertor(obj):
    """Convert object to be useable for json serialization.

    To be used in json.dumps as value for default= so that given object
    is serializable
    """
    if "Binary" in str(obj.__class__):
        return base64.b64encode(obj.data).decode("utf-8")

    if "DateTime" in str(obj.__class__):
        date = datetime.datetime.strptime(str(obj), "%Y%m%dT%H:%M:%S")
        return date.isoformat() + "Z"

    if "Decimal" in str(obj.__class__):
        return str(obj)

    # use string as default
    return str(obj)


def load_json(path):
    """Load JSON file into Python object."""
    with open(os.path.expanduser(path), "r", encoding="utf-8") as file_data:
        data = json.load(file_data)
    return data


def load_yaml(path):
    """Load YAML file into Python object."""
    with open(os.path.expanduser(path), "r", encoding="utf-8") as file_data:
        data = yaml.safe_load(file_data)
    return data


def save_to_json(path, data):
    """Serialize object into JSON file."""
    try:
        with open(os.path.expanduser(path), "w", encoding="utf-8") as output:
            json.dump(data, output, default=json_convertor, indent=2, sort_keys=True)
    except IOError as exc:
        logger.exception(exc)
        sys.exit(1)


@contextlib.contextmanager
def fd_open(filename=None):
    """Use file or stout as output file descriptor."""
    if filename:
        file_d = open(os.path.expanduser(filename), "w", encoding="utf-8")
    else:
        file_d = sys.stdout

    try:
        yield file_d
    finally:
        if file_d is not sys.stdout:
            file_d.close()


def save_yaml(path, yaml_data, **kwargs):
    """
    Write yaml data with file.write(yaml.dump()) to file specified by path.

    If path is not specified use stdout.
    """
    with fd_open(path) as yaml_file:
        yaml_file.write(yaml.dump(yaml_data, default_flow_style=False, **kwargs))


def object2json(obj):
    """Convert object to JSON string."""
    return json.dumps(obj, default=json_convertor, sort_keys=True, indent=4)


def print_obj(obj):
    """Print object as JSON."""
    logger.info(object2json(obj))


def get_host_from_metadata(metadata, name):
    """
    Get host definition from job metadata base on name.

    Returns:
    (host, domain)
    """
    domains = metadata.get("domains", [])
    for domain in domains:
        for host in domain.get("hosts", []):
            if host["name"] == name:
                return host, domain

    return None, None


def is_windows_host(meta_host):
    """
    Return if host is Windows host based on host metadata info.

    Host is windows host if:
    * os starts with 'win' or
    * os_type is 'windows'
    """
    return (
        meta_host.get("os", "").startswith("win")
        or meta_host.get("os_type", "") == "windows"
    )


def get_os_type(meta_host):
    """Get os type from ``os_type`` or infer it from ``os``."""
    os_type = meta_host.get("os_type")
    if not os_type:
        os_type = "linux" if not is_windows_host(meta_host) else "windows"

    return os_type


def get_shortname(hostname):
    """
    Get shortname part of fqdn.

    Return the same string if it is not fqdn.
    """
    return hostname.split(".")[0]


def get_fqdn(hostname, domain):
    """
    Get fqdn from hostname and domain.

    Return the hostname if it is fqdn otherwise append domain.
    """
    if "." in hostname:
        return hostname

    return f"{hostname}.{domain}"


def get_username(host, meta_host, config):
    """Find username from sources db/metadata/config."""
    provider = host.provider.name
    dict_key = meta_host["os"]
    default_user = None
    if is_windows_host(meta_host):
        default_user = "Administrator"
    username = find_value_in_config_hierarchy(
        config, provider, host, meta_host, "username", "users", dict_key, default_user
    )

    return username


def get_password(host, meta_host, config):
    """Find password from sources db/metadata/config."""
    password = host.password or meta_host.get("password")
    multihost_config = config.get("mhcfg")
    if is_windows_host(meta_host) and multihost_config:
        password = password or multihost_config.get("ad_admin_password")
    return password


def get_ssh_key(host, meta_host, config):
    """Find ssh_key path from sources: db/metadata/config."""
    ssh_key_attr = "ssh_key_filename"
    provider_config = config.get(host.provider.name, {})

    ssh_key = meta_host.get(ssh_key_attr)
    ssh_key = ssh_key or provider_config.get(ssh_key_attr)
    ssh_key = ssh_key or config.get(ssh_key_attr)
    return ssh_key


def get_username_pass_and_ssh_key(host, context):
    """Return username password and ssh_key to be later used for ssh connections."""
    meta_host, _domain = get_host_from_metadata(context.METADATA, host.name)
    username = get_username(host, meta_host, context.PROV_CONFIG)
    ssh_key = get_ssh_key(host, meta_host, context.PROV_CONFIG)
    password = None if ssh_key else get_password(host, meta_host, context.PROV_CONFIG)
    return username, password, ssh_key


def ssh_to_host(
    host,
    username=None,
    password=None,
    ssh_key=None,
    command=None,
    interactive=False,
):
    """SSH to the selected host."""
    psw = host.password or password

    run_args = {
        "env": os.environ.copy(),
        "shell": True,
    }
    if not interactive:
        run_args.update(
            {
                "stdout": subprocess.PIPE,
                "stderr": subprocess.PIPE,
            }
        )

    cmd = ["ssh"]
    cmd.extend(["-o", "'StrictHostKeyChecking=no'"])
    cmd.extend(["-o", "'UserKnownHostsFile=/dev/null'"])

    if psw:
        cmd.extend(["-o", "'PasswordAuthentication=yes'"])
        cmd = ["sshpass", "-p", psw] + cmd
    elif ssh_key:
        cmd.extend(["-o", "'PasswordAuthentication=no'"])
        cmd.extend(["-i", ssh_key])

    if username:
        cmd.extend(["-l", username])

    cmd.append(host.ip_addr)  # Destination

    if command:
        cmd.append(command)

    cmd = " ".join(cmd)

    logger.debug(f"Running: {cmd}")
    with subprocess.Popen(cmd, **run_args) as process:
        std_out, std_err = process.communicate()

    if not interactive:
        for o_line in std_out.decode().splitlines():
            logger.debug(f"stdout: {o_line}")

        for e_line in std_err.decode().splitlines():
            logger.debug(f"stderr: {e_line}")

    return process.returncode == 0


async def exec_async_subprocess(program, args, raise_on_err=True):
    """Util method to execute subprocess asynchronously."""
    process = await asyncio.create_subprocess_exec(
        program,
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()
    if stdout:
        stdout = stdout.decode()
    if stdout is None:
        stdout = ""
    if stderr:
        stderr = stderr.decode()
    if stdout is None:
        stderr = ""
    if process.returncode != 0 and raise_on_err:
        raise ProvisioningError(stderr)
    return stdout, stderr, process


def async_run(func):
    """Decorate click actions to run as async."""

    @wraps(func)
    def update_wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return update_wrapper


class NoSuchFileHandler:
    """
    Decorator which change error into ConfigError with custom message.

    error string can contain {path} template.
    """

    def __init__(self, error):
        """Initialize decorator."""
        self.error = error

    def __call__(self, func):
        """Transform FileNotFoundError into ConfigError."""

        def wrapped_f(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except FileNotFoundError as file_error:
                raise ConfigError(
                    self.error.format(path=file_error.filename)
                ) from file_error

        return wrapped_f



# Copyright 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Virt provider."""

import asyncio
import getpass
import grp
import logging
import os
from datetime import datetime

from testcloud.exceptions import TestcloudImageError

from mrack.errors import ProvisioningError, ValidationError
from mrack.host import STATUS_ACTIVE, STATUS_OTHER, STATUS_PENDING
from mrack.providers.provider import STRATEGY_ABORT, Provider
from mrack.providers.utils.testcloud import Testcloud
from mrack.utils import is_windows_host

logger = logging.getLogger(__name__)

PROVISIONER_KEY = "virt"


class VirtProvider(Provider):
    """Virt provisioning provider."""

    def __init__(self):
        """Initialize provider."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "Virt"
        self.testcloud = Testcloud()
        self.max_retry = 1  # for retry strategy
        self.status_map = {
            "running": STATUS_ACTIVE,
            "shutoff": STATUS_OTHER,
            "undefined": STATUS_OTHER,
            "de-sync": STATUS_PENDING,
        }

    def _extract_err_msg(self, virt_err):
        """Extract Virt provider traceback error from message."""
        # split traceback string to find Error causing failure
        # index 1 should point to the Error string
        exc_str = "\n".join([f"\t{s}" for s in str(virt_err).splitlines()])
        logger.debug(f"{self.dsp_name} Exception occured:\n{exc_str}")
        err_sep = "Error:"
        if err_sep in exc_str:
            err_str = str(virt_err).split(err_sep)[1].strip()
        elif err_sep.upper() in exc_str:
            err_str = str(virt_err).split(err_sep.upper())[1].strip()
        else:
            err_str = exc_str
        # return first line of error as it could be more lines of traceback
        return err_str.split("\n")[0]

    async def init(self, strategy=STRATEGY_ABORT, max_retry=1):
        """Initialize Virt provider with data from config."""
        logger.info(f"{self.dsp_name} Initializing provider")
        login_start = datetime.now()
        self.strategy = strategy
        self.max_retry = max_retry
        login_end = datetime.now()
        login_duration = login_end - login_start
        logger.info(f"{self.dsp_name} Init duration {login_duration}")

    async def prepare_provisioning(self, reqs):
        """Prepare provisioning."""
        # Check that current user belongs to testcloud group
        group = "testcloud"
        uname = getpass.getuser()
        _name, _pass, gid, _members = grp.getgrnam(group)
        groups = os.getgroups()
        if gid not in groups:
            raise ValidationError(
                f"Error: Current user is not a member of a {group} group.\n"
                "Make sure that testcloud is installed and add the user to the "
                "group, e.g. by:\n"
                f"  sudo usermod -a -G {group} {uname}"
            )

        # Pulling images ahead so that the provider doesn't download the same
        # image more than once
        pull = set()
        for req in reqs:
            pull.add(req["image_url"])

        awaitables = []
        for url in pull:
            logger.info(f"{self.dsp_name} Pulling image '{url}'")
            awaitables.append(self.testcloud.pull_image(url))

        pull_results = await asyncio.gather(*awaitables, return_exceptions=True)
        success = all(pull_results)

        for pull in pull_results:
            if isinstance(pull, TestcloudImageError):
                logger.error(f"{self.dsp_name} {str(pull)}")
                success = False
            elif isinstance(pull, Exception):
                raise pull

        if not success:
            logger.error(f"{self.dsp_name} Pulling of images failed")
        else:
            logger.info(f"{self.dsp_name} Images prepared")

        return success

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        return bool(reqs)  # no specific validation yet

    async def can_provision(self, hosts):
        """Check that provider has enough resources to provision hosts."""
        # We'd need to check available memory, this is TODO
        return bool(hosts)

    async def utilization(self):
        """Check percentage utilization of given provider."""
        return 0

    async def create_server(self, req):
        """Request and create resource on Virt provider."""
        hostname = req.get("name")
        logger.info(f"{self.dsp_name} [{hostname}] Creating virtual machine")

        host_id = req["run_id"] + "-" + hostname
        try:
            out, err, _proc = await self.testcloud.create(
                host_id,
                **req,
            )
        except ProvisioningError as virt_err:
            req.update({"host_id": host_id})
            raise ProvisioningError(self._extract_err_msg(virt_err), req) from virt_err

        info = self.testcloud.info(host_id)
        info["id"] = host_id
        info["name"] = hostname
        info["output"] = out
        if info["state"] != "running":
            info["error"] = err

        if is_windows_host(req):
            # testcloud can inject default password for the linux instances only so far
            del info["password"]

        return (info, req)

    async def wait_till_provisioned(self, resource):
        """Wait till resource is provisioned."""
        result, req = resource
        result.update({"mrack_req": req})
        return result, req

    async def delete_host(self, host_id, host_name):
        """Delete provisioned host."""
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        logger.info(f"{log_msg_start} Removing VM {host_id}")
        try:
            _out, _err, _proc = await self.testcloud.destroy(host_id)
        except ProvisioningError as p_err:
            # just log error message when unable to delete VM
            logger.error(f"{log_msg_start} {self._extract_err_msg(p_err)}")

        return True

    def prov_result_to_host_data(self, prov_result, req):
        """Get needed host information from podman provisioning result."""
        result = {}
        result["id"] = prov_result.get("id")
        result["name"] = prov_result.get("name")
        result["addresses"] = [prov_result.get("ip")]
        result["status"] = prov_result["state"]
        result["fault"] = prov_result.get("error")
        result["password"] = prov_result.get("password")
        result["os"] = prov_result.get("mrack_req").get("os")
        result["group"] = prov_result.get("mrack_req").get("group")

        return result



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenStack provider."""

import asyncio
import logging
from copy import deepcopy
from datetime import datetime, timedelta
from random import random, sample
from urllib.parse import parse_qs, urlparse

from asyncopenstackclient import AuthPassword, GlanceClient
from simple_rest_client.exceptions import NotFoundError, ServerError

from mrack.context import global_context
from mrack.errors import (
    NotAuthenticatedError,
    ProviderError,
    ProvisioningError,
    ServerNotFoundError,
    ValidationError,
)
from mrack.host import STATUS_ACTIVE, STATUS_DELETED, STATUS_ERROR, STATUS_PROVISIONING
from mrack.providers.provider import STRATEGY_ABORT, Provider
from mrack.providers.utils.osapi import ExtraNovaClient, NeutronClient
from mrack.utils import get_shortname, is_windows_host, object2json

logger = logging.getLogger(__name__)

# Docs
# https://github.com/DreamLab/AsyncOpenStackClient
# https://docs.openstack.org/queens/api/


PROVISIONER_KEY = "openstack"
SERVER_ERROR_RETRY = 5  # number of times to retry server creation
SERVER_ERROR_SLEEP = 10  # seconds
SERVER_RES_SLEEP = 10  # minutes
NETWORK_NAME = 0
NETWORK_SIZE = 1


class OpenStackProvider(Provider):
    """
    OpenStack Provider.

    Provisions servers in OpenStack with added logic to check if requested
    resources are available.
    """

    def __init__(self):
        """Object initialization."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "OpenStack"
        self.max_retry = 1  # provisioning retries
        self.flavors = {}
        self.flavors_by_ref = {}
        self.images = {}
        self.images_by_ref = {}
        self.limits = {}
        self.networks = {}
        self.networks_by_ref = {}
        self.ips = {}
        self.ips_by_ref = {}
        self.api_timeout = 6 * 60  # timeout for request to OpenStack
        self.timeout = 60  # minutes
        self.poll_sleep_initial = 15  # seconds
        self.poll_sleep = 7  # seconds
        self.poll_init_adj = 0  # set based on # of hosts to provisions
        self.poll_adj = 0  # set based on # of hosts to provisions
        self.status_map = {
            "ACTIVE": STATUS_ACTIVE,
            "BUILD": STATUS_PROVISIONING,
            "DELETED": STATUS_DELETED,
            "ERROR": STATUS_ERROR,
            # there is much more we can treat it as STATUS_OTHER, see:
            # https://docs.openstack.org/api-guide/compute/server_concepts.html
        }

    async def _openstack_gather_responses(self, *calls):
        """Gather the async result of functions from parameters.

        Returns:
        * list of asyncio.gather results
        """
        error_attempts = 0
        coros = [func(*args, **xargs) for (func, args, xargs) in calls]
        result = []
        while error_attempts < SERVER_ERROR_RETRY:
            try:
                result = await asyncio.gather(*coros)
                break
            except ServerError as exc:
                logger.debug(f"{self.dsp_name} {exc}")
                error_attempts += 1
                if error_attempts <= SERVER_ERROR_RETRY:
                    await asyncio.sleep(SERVER_ERROR_SLEEP)
                    # create new coroutines on ServerError
                    coros = [func(*args, **xargs) for (func, args, xargs) in calls]
                    continue  # Try again due to ServerError
        else:
            # now we are past to what we would like to wait fail now
            raise ProvisioningError(
                f"{self.dsp_name} Failed to load environment objects from server",
            )

        return result

    async def init(
        self, image_names=None, networks=None, strategy=STRATEGY_ABORT, max_retry=1
    ):
        """Initialize provider with data from OpenStack.

        Load:
        * available flavors
        * networks
        * network availabilities (number of available IPs for networks)
        * images which were defined in `images` option
        * account limits (max and current usage of vCPUs, memory, ...)
        """
        # session expects that credentials will be set via env variables
        logger.info(f"{self.dsp_name} Initializing provider")
        self.strategy = strategy
        self.max_retry = max_retry
        try:
            self.session = AuthPassword()
        except TypeError as terr:
            err = (
                "OpenStack credentials not provided. Load OpenStack RC file with valid "
                "credentials and try again. E.g.: $ source PROJECT-openrc.sh"
            )
            raise NotAuthenticatedError(err) from terr

        self.nova = ExtraNovaClient(session=self.session)
        self.glance = GlanceClient(session=self.session)
        self.neutron = NeutronClient(session=self.session)

        login_start = datetime.now()
        await asyncio.gather(
            self.nova.init_api(self.api_timeout),
            self.glance.init_api(self.api_timeout),
            self.neutron.init_api(self.api_timeout),
        )
        login_end = datetime.now()
        logger.info(f"{self.dsp_name} Login duration {login_end - login_start}")

        self.network_pools = networks
        object_start = datetime.now()

        _, _, self.limits, _, _ = await self._openstack_gather_responses(
            [self._load_flavors, [], {}],
            [self._load_images, [image_names], {}],
            [self.nova.limits.show, [], {}],
            [self._load_networks, [], {}],
            [self._load_ip_availabilities, [], {}],
        )

        object_duration = datetime.now() - object_start
        logger.info(
            f"{self.dsp_name} Environment objects load duration: {object_duration}"
        )

    def _set_flavors(self, flavors):
        """Extend provider configuration with list of flavors."""
        for flavor in flavors:
            self.flavors[flavor["name"]] = flavor
            self.flavors_by_ref[flavor["id"]] = flavor

    def _set_images(self, images):
        """Extend provider configuration with list of images."""
        for image in images:
            self.images[image["name"]] = image
            self.images_by_ref[image["id"]] = image

    def _is_network_type(self, name):
        """Check if name is a configured network type in provisioning config."""
        network_type = self.network_pools.get(name)
        return bool(network_type)

    def _aggregate_networks(self, hosts):
        """
        Get how many host require each used network type.

        Returns: dict where keys are network types and values are total count.
        """
        network_types = {}
        for host in hosts:
            # skip hosts which have low-level network names defined
            # this can be extended to pick network type based on the network name
            names = host.get("networks")
            if names:
                continue
            network_type = host.get("network")
            if not self._is_network_type(network_type):
                continue

            count = network_types.get(network_type, 0)
            count += 1
            network_types[network_type] = count

        return network_types

    def _pick_network(self, host_name, host_weight, nets_weights):
        """
        Pick network based network type and weight of host.

        This method allows OpenStack to pick Network from
        all of networks based on network load and requirement
        got from the metadata. Each of required host gets
        its own position, aka weigh in interval <0, 1>
        based on the order (index) in metadata. Before that
        all of the networks are 'normalized' with _get_weights_from_usable

        An example:
        - considering 5 host request
        - picked 5 networks where availability is > 5%

        Host weights:
        - host 1 - 0/5 = 0
        - host 2 - 1/5 = 0.2
        - host 3 - 2/5 = 0.4
        - host 4 - 3/5 = 0.6
        - host 5 - 4/5 = 0.8

        Network availability:
        - net 1 - 20 addresses
        - net 2 - 100 addresses
        - net 3 - 130 addresses
        - net 4 - 145 addresses
        - net 5 - 105 addresses

        Full capacity of these 5 nets: 500

        Normalized network range:
        net 4 - <0, 0.29)
        net 3 - <0,29, 0.55)
        net 5 - <0.55, 0.76)
        net 2 - <0.76, 0.96)
        net 1 - <0.96, 1)

        Which will divide networks for hosts:
        host 1 => net 4 (0 falls into net 4 interval)
        host 2 => net 4 (0.2 falls into net 4 interval)
        host 3 => net 3 (0.4 falls into net 3 interval)
        host 4 => net 5 (0.6 falls into net 5 interval)
        host 5 => net 2 (0.8 falls into net 2 interval)

        net 1 will be unused as bigger networks are preferred this way.
        """
        chosen = tuple()
        for index, net in enumerate(nets_weights):
            if host_weight < net[NETWORK_SIZE]:
                chosen = net
                # print details only if there is more one network to be picked from
                if len(nets_weights) > 1:
                    lower_bound = nets_weights[index - 1][NETWORK_SIZE] if index else 0
                    logger.debug(
                        f"{self.dsp_name} [{host_name}] Weight of host "
                        f"{host_weight:.3f} fell into interval: "
                        f"({lower_bound:.3f}, {chosen[NETWORK_SIZE]:.3f}>"
                    )

                break

        if not chosen:
            raise ProviderError(f"{self.dsp_name} Error: No network has been chosen")

        logger.debug(
            f"{self.dsp_name} [{host_name}] Picked network '{chosen[NETWORK_NAME]}'"
        )

        return chosen[NETWORK_NAME]

    def _get_usable_networks(self, network_type, requested_ip_cnt):
        """
        Return list of available networks and ip availability.

        Available networks with most IPs available is picked.
        """
        usable = set()
        low_avail_nets = set()
        total_available = 0
        max_size_net = 0
        threshold = global_context.CONFIG.usable_network_threshold
        spread_option = global_context.CONFIG.network_spread
        big_request = False

        # filter out the None values - the get_network return None if network not found
        for network in list(
            filter(
                None,
                [self._get_network(net) for net in self.network_pools[network_type]],
            )
        ):
            ips = self._get_ips(ref=network.get("id"))

            available = ips["total_ips"] - ips["used_ips"]
            max_size_net = available if available > max_size_net else max_size_net

            logger.debug(
                f"{self.dsp_name} Network: {network['name']}"
                f"{' - unusable (0 IPs left) skipping' if not available else ''}"
            )
            logger.debug(f"  total: {ips['total_ips']}")
            logger.debug(f"  used: {ips['used_ips']}")
            logger.debug(f"  available: {available}")

            if not available:
                continue

            total_available += available

            #  assume network is unusable when usable_threshold is reached
            if ips["used_ips"] / ips["total_ips"] > threshold / 100:
                low_avail_nets.add((network["name"], available))
                continue

            usable.add((network["name"], available))

        if not usable and (total_available >= requested_ip_cnt) and low_avail_nets:
            usable = usable.union(low_avail_nets)

        if not usable:
            raise ValidationError(
                "No available networks for "
                f"{requested_ip_cnt} hosts with {network_type}",
                self.dsp_name,
            )

        if requested_ip_cnt > max_size_net:
            big_request = True
            # if we request more IPs than maximum size net consider the request big
            # adding even small networks to the usable set to spread the load
            usable = usable.union(low_avail_nets)

        logger.debug(
            f"{self.dsp_name} Default mrack behavior is to 'allow' spread"
            " the load to all available networks, to disable this feature set"
            " the 'network-spread' option in the mrack.conf file to 'no'"
        )
        logger.debug(
            f"{self.dsp_name} 'network-spread' option is set to '{spread_option}'"
        )
        logger.debug(
            f"{self.dsp_name} Considering this request "
            f"{'big' if big_request else 'small'} - requesting {requested_ip_cnt} "
            f"IP(s) from total available {total_available} IP(s)"
        )

        size_for_spread = (
            len(usable) if len(usable) <= requested_ip_cnt else requested_ip_cnt
        )

        # number of networks which we will consider to pass to picking algorithm
        network_subset_size = 0

        if spread_option == "force":
            network_subset_size = size_for_spread
        elif spread_option == "no":
            if not big_request:
                network_subset_size = 1
        else:  # spread option is allow or some unknown string -> fallback to default
            if usable.intersection(low_avail_nets) or big_request:
                network_subset_size = size_for_spread
            else:
                network_subset_size = 1

        if network_subset_size:
            return sorted(
                set(sample(sorted(usable), network_subset_size)),
                key=lambda u: u[NETWORK_SIZE],
                reverse=True,
            )
        # If the above return does not happened because there is 0 usable networks
        # we will raise a ValidationError
        raise ValidationError(
            f"Can not satisfy request for {requested_ip_cnt} hosts ({network_type}) "
            "Change the 'network-spread' in mrack.conf or try provisioning again later",
            self.dsp_name,
        )

    def _get_weights_from_usable(self, usable_networks):
        """
        Define map of network type and weight of usable_networks.

        All of the networks are 'normalized' in a way that the random
        selection from subset of networks (networks_weights parameter)
        (or all of network if number of hosts >= number of networks)
        is divided to effective range (multiple intervals in <0, 1>)
        for the network based on the relative network weight
        compared to full capacity of considered networks.

        e.g.:
        - usable_networks = [('net_1', 11), ('net_3', 10)]
        - weights = [('net_1', 0.5238095238095238), ('net_3', 1.0)]

        """
        total_weight = sum(net[NETWORK_SIZE] for net in usable_networks)
        weights = [
            (net[NETWORK_NAME], net[NETWORK_SIZE] / total_weight)
            for net in usable_networks
        ]
        if len(weights) == 1:
            return weights

        for w_index in range(1, len(weights)):
            weights[w_index] = (
                weights[w_index][NETWORK_NAME],
                weights[w_index - 1][NETWORK_SIZE] + weights[w_index][NETWORK_SIZE],
            )

        return weights

    def _translate_network_types(self, hosts):
        """Pick the right OpenStack networks for all hosts.

        Pick the network based on network type, networks configured for the
        type and the available IP addresses. Process all hosts to
        be able to pick the network which have enough addresses for all hosts.

        All hosts will have either "networks" attribute or "network"
        host attribute set with OpenStack network name or ID.
        """
        nt_requirements = self._aggregate_networks(hosts)
        nt_map = {}
        for network_type, count in nt_requirements.items():
            # count is crucial
            nt_map[network_type] = self._get_usable_networks(network_type, count)

        weight_map = {}
        for network_type, usable_networks in nt_map.items():
            weight_map[network_type] = self._get_weights_from_usable(
                usable_networks,
            )

        for net_type, netw_weights in weight_map.items():
            logger.debug(
                f"{self.dsp_name} Considered networks "
                f"for {net_type} (network_name, weight_interval):"
            )
            for net in netw_weights:
                logger.debug(f"  {net}")

        for host in hosts:
            # skip hosts which have low-level network names defined
            names = host.get("networks")
            if names:
                continue

            network_type = host.get("network")

            # skip if network_type is not network type
            if not self.network_pools.get(network_type):
                continue

            networks_weights = weight_map[network_type]

            host["network"] = self._pick_network(
                host_name=host["name"],
                host_weight=hosts.index(host) / len(hosts),
                nets_weights=networks_weights,
            )

    def _set_networks(self, networks):
        """Extend provider configuration with list of networks."""
        for network in networks:
            self.networks[network["name"]] = network
            self.networks_by_ref[network["id"]] = network

    def _get_flavor(self, name=None, ref=None):
        """Get flavor by name or UUID."""
        flavor = self.flavors.get(name)
        if not flavor:
            flavor = self.flavors_by_ref.get(ref)
        return flavor

    def _get_image(self, name=None, ref=None):
        """Get image by name or UUID."""
        image = self.images.get(name)
        if not image:
            image = self.images_by_ref.get(ref)
        return image

    def _get_network(self, name=None, ref=None):
        """Get network by name or UUID."""
        network = self.networks.get(name)
        if not network:
            network = self.networks_by_ref.get(ref)

        if not network:
            net_id = name or ref
            logger.debug(
                f"{self.dsp_name} Failed to load network with name: '{net_id}'"
            )

        return network

    def _get_ips(self, name=None, ref=None):
        """Get network availability by network name or network UUID."""
        aval = self.ips.get(name)
        if not aval:
            aval = self.ips_by_ref.get(ref)
        return aval

    async def _load_flavors(self):
        """Extend provider configuration by loading all flavors from OpenStack."""
        resp = await self.nova.flavors.list()
        flavors = resp["flavors"]
        self._set_flavors(flavors)
        return flavors

    async def _load_images(self, image_names=None):
        """
        Extend provider configuration by loading information about images.

        Load everything if image_names list is not specified.

        Specifying list of images to load might improve performance if the
        OpenStack instance contains a lot of images.
        """
        params = {"limit": 1000}

        if image_names:
            image_filter = ",".join(image_names)
            image_filter = "in:" + image_filter
            params["name"] = image_filter

        images = []
        response = await self.glance.images.list(**params)
        images.extend(response["images"])

        while response.get("next"):
            p_result = urlparse(response.get("next"))
            query = p_result.query
            next_params = parse_qs(query)
            for key, val in next_params.items():
                if isinstance(val, list) and val:
                    next_params[key] = val[0]
            response = await self.glance.images.list(**next_params)
            images.extend(response["images"])

        self._set_images(images)

        return images

    async def _load_networks(self):
        """Extend provider configuration by loading all networks from OpenStack."""
        resp = await self.neutron.network.list()
        networks = resp["networks"]
        self._set_networks(networks)
        return networks

    async def _load_ip_availabilities(self):
        """Extend provider configuration by loading networks availabilities."""
        resp = await self.neutron.ip.list()
        availabilities = resp["network_ip_availabilities"]
        for availability in availabilities:
            self.ips[availability["network_name"]] = availability
            self.ips_by_ref[availability["network_id"]] = availability
        return availabilities

    def _translate_flavor(self, req):
        flavor_spec = req.get("flavor")
        flavor_ref = req.get("flavorRef")
        flavor = None
        if flavor_ref:
            flavor = self._get_flavor(ref=flavor_ref)
        if flavor_spec:
            flavor = self._get_flavor(flavor_spec, flavor_spec)

        if not flavor:
            specs = f"flavor: {flavor_spec}, ref: {flavor_ref}"
            raise ValidationError(f"Flavor not found: {specs}", self.dsp_name)
        return flavor

    def _translate_image(self, req):
        image_spec = req.get("image")
        image_ref = req.get("imageRef")
        image = None
        if image_ref:
            image = self._get_image(ref=image_ref)
        if image_spec:
            image = self._get_image(image_spec, image_spec)
        if not image:
            specs = f"image: {image_spec}, ref: {image_ref}"
            raise ValidationError(f"Image not found {specs}", self.dsp_name)
        return image

    def _translate_networks(self, req, spec=False):
        network_req = req.get("network")
        network_specs = req.get("networks", [])
        network_specs = deepcopy(network_specs)
        networks = []
        if not isinstance(network_specs, list):
            network_specs = []
        for network_spec in network_specs:
            uuid = network_spec.get("uuid")
            network = self._get_network(ref=uuid)
            if not network:
                raise ValidationError(
                    f"Network not found: {network_spec}", self.dsp_name
                )
            networks.append(network)
        if network_req:
            network = self._get_network(name=network_req, ref=network_req)
            if not network:
                raise ValidationError(
                    f"Network not found: {network_req}", self.dsp_name
                )

            network_specs.append({"uuid": network["id"]})
            networks.append(network)

        if spec:
            return network_specs
        return networks

    def validate_host(self, req):
        """Validate that host requirements contains existing required objects."""
        self._translate_flavor(req)
        self._translate_image(req)
        self._translate_networks(req)

        return True

    def _set_poll_sleep_times(self, reqs):
        """
        Compute polling sleep times based on number of hosts.

        So that we don't create unnecessary load on server while still checking returns
        (initial_sleep, sleep)
        """
        count = len(reqs)

        # initial poll is the biggest performance saver it should be around
        # time when more than half of host is in ACTIVE state
        self.poll_init_adj = 0.65 * count

        # poll time should ask often enough, to not create unnecessary delays
        # while not that many to not load the server much
        self.poll_adj = 0.22 * count

    async def prepare_provisioning(self, reqs):
        """
        Prepare provisioning.

        Load missing images if they are not in provisioning-config.yaml
        """
        prepare_images = list(
            {req["image"] for req in reqs if req["image"] not in self.images}
        )

        if prepare_images:
            im_list = ", ".join(prepare_images)
            logger.debug(f"{self.dsp_name} Loading image info for: '{im_list}'")
            await self._load_images(list(prepare_images))
            logger.debug(f"{self.dsp_name} Loading images info done.")

        self._set_poll_sleep_times(reqs)
        return True

    async def validate_hosts(self, reqs):
        """Validate that all hosts requirements contains existing required objects."""
        # translate network type to actual network and check network availabilities
        self._translate_network_types(reqs)

        for req in reqs:
            logger.info(f"{self.dsp_name} Validating host: {object2json(req)}")
            self.validate_host(req)
            logger.info(f"{self.dsp_name} [{req['name']}] OK")

    def get_host_requirements(self, req):
        """Get vCPU and memory requirements for host requirement."""
        flavor_spec = req.get("flavor")
        flavor_ref = req.get("flavorRef")
        flavor = None
        if flavor_ref:
            flavor = self._get_flavor(ref=flavor_ref)
        if flavor_spec:
            flavor = self._get_flavor(flavor_spec, flavor_spec)

        try:
            res = {"ram": flavor["ram"], "vcpus": flavor["vcpus"]}
        except TypeError as flavor_none:
            # func does not load flavor so None is used as result
            raise ValidationError(
                f"Could not load the flavor for requirement: {req}", self.dsp_name
            ) from flavor_none

        return res

    async def _load_limits(self):
        limits_await = await self._openstack_gather_responses(
            [self.nova.limits.show, [], {}],
        )
        self.limits = limits_await[0]  # gather returns list

        limits = self.limits["limits"]["absolute"]
        used_vcpus = limits["totalCoresUsed"]
        used_memory = limits["totalRAMUsed"]
        limit_vcpus = limits["maxTotalCores"]
        limit_memory = limits["maxTotalRAMSize"]

        return used_vcpus, used_memory, limit_vcpus, limit_memory

    async def can_provision(self, hosts):  # pylint: disable=arguments-differ
        """Check that all host can be provisioned.

        Checks:
        * available vCPUs and memory based on account limits
        * that all host contain available flavors, images, networks
        """
        vcpus = 0
        ram = 0

        for req in hosts:
            needs = self.get_host_requirements(req)
            vcpus += needs["vcpus"]
            ram += needs["ram"]

        # poll the actual openstack load
        logger.debug(f"{self.dsp_name} Loading nova limits")

        used_vcpus, used_memory, limit_vcpus, limit_memory = await self._load_limits()

        req_vcpus = used_vcpus + vcpus
        req_memory = used_memory + ram

        logger.info(
            f"{self.dsp_name} Required vcpus: {vcpus}, "
            f"used: {used_vcpus}, max: {limit_vcpus}"
        )

        logger.info(
            f"{self.dsp_name} Required ram: {ram}, "
            f"used: {used_memory}, max: {limit_memory}"
        )

        return req_vcpus <= limit_vcpus and req_memory <= limit_memory

    async def utilization(self):
        """Check utilization of provider."""
        used_vcpus, used_memory, limit_vcpus, limit_memory = await self._load_limits()
        cpu_util = used_vcpus / limit_vcpus * 100
        memory_util = used_memory / limit_memory * 100
        return cpu_util if memory_util <= cpu_util else memory_util

    async def create_server(self, req):
        """Issue creation of a server.

        req - dict of server requirements - can contains values defined in
              POST /servers official OpenStack API
              https://docs.openstack.org/api-ref/compute/?expanded=create-server-detail#create-server

        The req object can contain following additional attributes:
        * 'flavor': uuid or name of flavor to use
        * 'network': uuid or name of network to use. Will be added to networks
                     list if present
        """
        name = req.get("name")
        log_msg_start = f"{self.dsp_name} [{name}]"
        logger.info(f"{log_msg_start} Creating server")
        specs = deepcopy(req)  # work with own copy, do not modify the input
        del specs["os"]  # do not pass this to openstack

        if is_windows_host(req):
            # Windows support only shortnames for Cloudbase-Init - it is
            # derived from openstack vm name.
            specs["name"] = get_shortname(name)

        flavor = self._translate_flavor(req)
        specs["flavorRef"] = flavor["id"]
        if specs.get("flavor"):
            del specs["flavor"]

        image = self._translate_image(req)
        if image.get("meta_compose_id") and image.get("meta_compose_url"):
            logger.info(
                f"{log_msg_start} Image meta_compose_id: {image['meta_compose_id']}"
                f"\n{log_msg_start} Image meta_compose_url:"
                f" {image['meta_compose_url']}"
            )

        specs["imageRef"] = image["id"]
        if specs.get("image"):
            del specs["image"]

        network_specs = self._translate_networks(req, spec=True)
        specs["networks"] = network_specs
        if specs.get("network"):
            del specs["network"]

        if specs.get("group"):
            del specs["group"]

        error_attempts = 0
        while error_attempts < SERVER_ERROR_RETRY:
            try:
                response = await self.nova.servers.create(server=specs)
            except ServerError as exc:
                logger.debug(f"{log_msg_start} {exc}")
                error_attempts += 1
                if error_attempts <= SERVER_ERROR_RETRY:
                    await asyncio.sleep(SERVER_ERROR_SLEEP)
                    continue  # Try again due to ServerError

            fault = response["server"].get("fault", {})

            if fault.get("code") == 500:
                # In such scenario OpenStack might run out of hosts to provision
                # This is not related to reaching OpenStack quota but to OpenStack
                # itself being fully loaded and without free resources to provide
                logger.info(
                    f"{log_msg_start} Unable to allocate resources for the required "
                    f"server (all available resources busy)"
                )
                error_attempts += 1
                logger.info(
                    f"{log_msg_start} Retrying request in {SERVER_RES_SLEEP} minutes"
                )
                # We should wait for OpenStack for reasonable time to try to reprovision
                # This sleep time should be longer for higher probability for Openstack
                # having freed some resources for us even when we are not reaching quota
                await asyncio.sleep(SERVER_RES_SLEEP * 60)  # * 60 - sleep for minutes
            else:
                # provisioning seems to pass correctly break to return result
                break

        else:
            # now we are past to what we would like to wait fail now
            raise ProvisioningError(
                f"{log_msg_start} Failed to create server",
                req,  # add the requirement dictionary to traceback for later
            )

        return (response.get("server"), req)

    async def delete_server(self, uuid):
        """Issue deletion of server.

        Doesn't wait for the deletion to happen.
        """
        error_attempts = 0
        while True:
            try:
                await self.nova.servers.force_delete(uuid)
            except ServerError as exc:
                logger.debug(exc)
                error_attempts += 1
                if error_attempts > SERVER_ERROR_RETRY:
                    raise ProviderError(uuid) from exc
                await asyncio.sleep(SERVER_ERROR_SLEEP)
            except NotFoundError:
                logger.warning(
                    f"{self.dsp_name} Server with ID '{uuid}' not found, "
                    "probably already deleted"
                )
                break

    async def wait_till_provisioned(self, resource):  # pylint: disable=too-many-locals
        """
        Wait till server is provisioned.

        Provisioned means that server is in ACTIVE or ERROR state

        State is checked by polling. Polling can be controller via `poll_sleep` and
        `poll_sleep_initial` options. This is useful when provisioning a lot of
        machines as it is better to increase initial poll to not ask to often as
        provisioning resources takes some time.

        Waits till timeout happens. Timeout can be either specified or default provider
        timeout is used.

        Return information about provisioned server.
        """
        resource, req = resource
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"
        uuid = resource.get("id")

        poll_sleep_initial = self.poll_sleep_initial + self.poll_init_adj
        poll_sleep_initial = (
            poll_sleep_initial / 2 + poll_sleep_initial * random() * 1.5
        )
        poll_sleep = self.poll_sleep + self.poll_adj

        start = datetime.now()
        timeout_time = start + timedelta(minutes=self.timeout)

        # do not check the state immediately, it will take some time
        logger.debug(
            f"{log_msg_start} ID {uuid}: sleeping for {poll_sleep_initial:.1f} seconds"
        )
        await asyncio.sleep(poll_sleep_initial)

        resp = {}
        logger.debug(f"{log_msg_start} ID {uuid}: Waiting for host creation")
        error_attempts = 0
        while datetime.now() < timeout_time:
            try:
                resp = await self.nova.servers.get(uuid)
            except NotFoundError as nf_err:
                raise ServerNotFoundError(uuid) from nf_err
            except ServerError as err:
                logger.debug(f"{log_msg_start} {err}")
                error_attempts += 1
                if error_attempts > SERVER_ERROR_RETRY:
                    raise ProvisioningError(uuid) from err

            server = resp["server"]
            if server["status"] in ["ACTIVE", "ERROR"]:
                break

            poll_sleep += 0.5  # increase delays to check the longer it takes
            logger.debug(
                f"{log_msg_start} ID {uuid}: sleeping for {poll_sleep:.1f} seconds"
            )
            await asyncio.sleep(poll_sleep)

        done_time = datetime.now()
        prov_duration = (done_time - start).total_seconds()

        if datetime.now() >= timeout_time:
            logger.warning(
                f"{log_msg_start} ID {uuid}: host was not provisioned "
                f"within a timeout of {self.timeout} mins"
            )
        else:
            logger.info(
                f"{log_msg_start} ID {uuid}: host "
                f"was provisioned in {prov_duration:.1f}s"
            )

        server.update({"mrack_req": req})

        return server, req

    async def delete_host(self, host_id, host_name):
        """Issue deletion of host(server) from OpenStack."""
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        logger.info(f"{log_msg_start} Deleting host with ID {host_id}")
        await self.delete_server(host_id)
        return True

    def prov_result_to_host_data(self, prov_result, req):
        """Get needed host information from openstack provisioning result."""
        result = {}
        meta_extra = {}
        image = self._translate_image(req)
        # Check if these fields exists, not all images have them
        if image.get("meta_compose_id"):
            meta_extra["meta_compose_id"] = image.get("meta_compose_id")
        if image.get("meta_compose_url"):
            meta_extra["meta_compose_url"] = image.get("meta_compose_url")

        result["id"] = prov_result.get("id")
        result["name"] = req.get("name")
        networks = prov_result.get("addresses", {})
        result["addresses"] = [ip.get("addr") for n in networks.values() for ip in n]
        result["fault"] = prov_result.get("fault")
        result["status"] = prov_result.get("status")
        result["os"] = prov_result.get("mrack_req").get("os")
        result["group"] = prov_result.get("mrack_req").get("group")
        result["meta_extra"] = meta_extra

        return result



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provisioning transformers.

Transformers take job definitions, combine it with provisioning configuration
and returns provisioning requirements (input for provisioners).
"""

import logging
from typing import Set, Tuple, Type

from mrack.errors import ProviderNotExists
from mrack.transformers.transformer import Transformer

NAME = 0
CLASS = 1
installed_transformers: Set[Tuple[str, Type[Transformer]]] = set()
IMPORT_ERR_TEMPLATE = "Transformer '%s' not installed, skipping registration"
logger = logging.getLogger(__name__)

try:
    from mrack.transformers.aws import CONFIG_KEY as AWS_KEY
    from mrack.transformers.aws import AWSTransformer

    installed_transformers.add((AWS_KEY, AWSTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)

try:
    from mrack.transformers.beaker import CONFIG_KEY as BEAKER_KEY
    from mrack.transformers.beaker import BeakerTransformer

    installed_transformers.add((BEAKER_KEY, BeakerTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)

try:
    from mrack.transformers.openstack import CONFIG_KEY as OPENSTACK_KEY
    from mrack.transformers.openstack import OpenStackTransformer

    installed_transformers.add((OPENSTACK_KEY, OpenStackTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)

try:
    from mrack.transformers.podman import CONFIG_KEY as PODMAN_KEY
    from mrack.transformers.podman import PodmanTransformer

    installed_transformers.add((PODMAN_KEY, PodmanTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)

try:
    from mrack.transformers.static import CONFIG_KEY as STATIC_KEY
    from mrack.transformers.static import StaticTransformer

    installed_transformers.add((STATIC_KEY, StaticTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)

try:
    from mrack.transformers.virt import CONFIG_KEY as VIRT_KEY
    from mrack.transformers.virt import VirtTransformer

    installed_transformers.add((VIRT_KEY, VirtTransformer))
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR_TEMPLATE, import_err.name)


class Registry:
    """Transformers registry."""

    def __init__(self):
        """Initialize transformer registry."""
        self._transformer_cls = {}
        self._transformers = {}

    def register(self, name, transformer_cls):
        """Register new tranformer class."""
        self._transformer_cls[name] = transformer_cls

    def get(self, name):
        """Get Transformer instance by name.

        Initialize the the transformer if not done yet.
        """
        trans_cls = self._transformer_cls.get(name)
        if not trans_cls:
            raise ProviderNotExists(f"Transformer '{name}' doesn't exist")
        transformer = self._transformers.get(name)
        if not transformer:
            transformer = trans_cls()
            self._transformers[name] = transformer
        return transformer

    @property
    def names(self):
        """Get all registered transformer names."""
        return self._transformer_cls.keys()


transformers = Registry()

for installed_transformer in installed_transformers:
    transformers.register(
        installed_transformer[NAME],
        installed_transformer[CLASS],
    )



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Destroy action module."""

import asyncio
import logging

from mrack.actions.action import Action
from mrack.host import STATUS_DELETED

logger = logging.getLogger(__name__)


class Destroy(Action):
    """Destroy action.

    Destroy all still active provisioned host. Save the state to DB.
    """

    async def destroy(self):
        """Execute the destroy action."""
        hosts = self._db_driver.hosts.values()
        to_del = [host for host in hosts if host.status != STATUS_DELETED]
        names = [host.name for host in to_del]
        names = ", ".join(names)
        logger.info(f"Hosts to delete: {names}")

        await self.init_providers(to_del)
        results_aws = []

        for host in to_del:
            logger.info(f"Deleting host: {host}")
            awaitable = host.delete()
            results_aws.append(awaitable)
        delete_results = await asyncio.gather(*results_aws)
        success = all(delete_results)
        self._db_driver.update_hosts(hosts)
        logger.info("Destroy done")
        return success

    async def init_providers(self, hosts):
        """Initialize providers for hosts to delete."""
        providers = [host.provider.name for host in hosts]
        providers = set(providers)
        aws = [self._get_transformer(provider) for provider in providers]
        await asyncio.gather(*aws)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenStack transformer module."""
import logging

from mrack.errors import ProvisioningConfigError
from mrack.providers.provider import STRATEGY_ABORT
from mrack.transformers.transformer import Transformer

logger = logging.getLogger(__name__)

CONFIG_KEY = "openstack"
DEFAULT_ATTEMPTS = 5


class OpenStackTransformer(Transformer):
    """OpenStack transformer."""

    _config_key = CONFIG_KEY
    _required_config_attrs = ["flavors", "networks", "images", "keypair"]  # List[str]

    async def init_provider(self):
        """Initialize associate provider and transformer display name."""
        self.dsp_name = "OpenStack"
        await self._provider.init(
            image_names=self.config["images"].values(),
            networks=self.config["networks"],
            strategy=self.config.get("strategy", STRATEGY_ABORT),
            max_retry=self.config.get("max_retry", DEFAULT_ATTEMPTS),
        )

    def _get_network_type(self, host):
        """Get network type from host object definition.

        If host object doesn't have it defined, then get it from metadata or
        provisioning config.
        """
        network_type = host.get("network")
        default_network = self.config.get("default_network")
        if network_type is None:
            network_type = self._metadata.get("network", default_network)
        if not network_type:
            raise ProvisioningConfigError(
                "No network type specified and project doesn't have default "
                "network type (property 'default_network') specified in "
                "provisioning config."
            )
        return network_type

    def validate_host(self, host):
        """Validate host input that it contains what OpenStack needs."""
        super().validate_host(host)
        self.validate_ownership_and_lifetime(host)

    def create_host_requirement(self, host):
        """Create single input for OpenStack provisioner."""
        req = {
            "name": host["name"],
            "os": host["os"],
            "group": host["group"],
            "flavor": self._get_flavor(host),
            "image": self._get_image(host),
            "key_name": self.config["keypair"],
            "network": self._get_network_type(host),
        }

        # do not add this requirement at all if it is not required
        config_drive_req = self.config.get("enable_config_drive")
        if config_drive_req:
            req.update({"config_drive": config_drive_req})

        req = self.update_metadata_for_owner_lifetime(req)
        return req



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""File database driver module."""

from os import path

from mrack.host import host_from_json
from mrack.utils import load_json, save_to_json

HOSTS_KEY = "hosts"


class FileDBDriver:
    """File database driver.

    Serialize and load information into JSON file.
    """

    def __init__(self, file_path):
        """Initialize DB driver."""
        self._path = file_path
        self._hosts = {}
        self._raw_data = None
        self.save_on_change = True
        self.load()

    def load(self):
        """Load configuration from file."""
        self._hosts = {}
        if not path.exists(self._path):
            self._raw_data = {HOSTS_KEY: {}}
            return self._hosts

        self._raw_data = load_json(self._path)
        raw_hosts = self._raw_data.get(HOSTS_KEY, [])

        self._hosts = {}
        for raw_host in raw_hosts:
            host = host_from_json(raw_host)
            self._hosts[host.name] = host

        return self._hosts

    def save(self):
        """Save configuration to file."""
        hosts = [host.to_json() for host in self._hosts.values()]
        self._raw_data[HOSTS_KEY] = hosts
        save_to_json(self._path, self._raw_data)

    @property
    def hosts(self):
        """Get all host objects loaded or to be saved."""
        return self._hosts

    def add_hosts(self, hosts):
        """Add a host object.

        Save it to file automatically if `save_on_change` is set to True.
        """
        for host in hosts:
            self.hosts[host.name] = host

        if self.save_on_change:
            self.save()

    def update_hosts(self, hosts):
        """Update managed host objects.

        Only adds.
        """
        self.add_hosts(hosts)

    def delete_host(self, host):
        """Delete host object."""
        if host.name in self.hosts:
            del self.hosts[host.name]



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Beaker Provider interface."""

import asyncio
import logging
import os
import socket
import xml.etree.ElementTree as eTree
from copy import deepcopy
from datetime import datetime, timedelta
from xml.dom.minidom import Document as xml_doc
from xmlrpc.client import Fault

from bkr.client import BeakerJob, BeakerRecipe, BeakerRecipeSet
from bkr.common.hub import HubProxy
from bkr.common.pyconfig import PyConfigParser
from gssapi.exceptions import MissingCredentialsError
from gssapi.raw.misc import GSSError

from mrack.errors import NotAuthenticatedError, ProvisioningError, ValidationError
from mrack.host import (
    STATUS_ACTIVE,
    STATUS_DELETED,
    STATUS_ERROR,
    STATUS_OTHER,
    STATUS_PROVISIONING,
)
from mrack.providers.provider import STRATEGY_ABORT, Provider
from mrack.utils import add_dict_to_node

logger = logging.getLogger(__name__)

PROVISIONER_KEY = "beaker"


def parse_bkr_exc_str(exc_str):
    """Parse exception string and return response dictionary for mrack error."""
    # we expect exception string to look like following:
    # '<class \'bkr.common.bexceptions.BX\'>:No distro tree matches Recipe:
    # <distroRequires>
    #   <and>
    #     <distro_name op="like" value="Fedora-33%"/>
    #   </and>
    # </distroRequires>'
    if (
        ":" not in exc_str.faultString
        and "bkr.common.bexceptions" not in exc_str.faultString
    ):
        # we got string we do not expect so just use the traceback
        return {"response": str(exc_str)}

    # because of expected format we split by ":" and use last 2 values from list
    # in above example it would be
    # [
    #   '\tNo distro tree matches Recipe',
    #   '\t<distroRequires><and><distro_name op="like" value="Fedora-33%"/> ...
    # ]
    fault = [f"\t{f.strip()}" for f in exc_str.faultString.split(":")[-2:]]
    return {"response": "\n".join(fault)}


class BeakerProvider(Provider):
    """Beaker Provider."""

    def __init__(self):
        """Object initialization."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "Beaker"
        self.conf = PyConfigParser()
        self.poll_sleep = 45  # seconds
        self.max_retry = 1  # for retry strategy
        self.status_map = {
            "Reserved": STATUS_ACTIVE,
            "New": STATUS_PROVISIONING,
            "Scheduled": STATUS_PROVISIONING,
            "Queued": STATUS_PROVISIONING,
            "Processed": STATUS_PROVISIONING,
            "Waiting": STATUS_PROVISIONING,
            "Installing": STATUS_PROVISIONING,
            "Running": STATUS_PROVISIONING,
            "Cancelled": STATUS_DELETED,
            "Aborted": STATUS_ERROR,
            "Completed": STATUS_OTHER,
            "MRACK_REACHED_TIMEOUT": STATUS_ERROR,
            "MRACK_RESULT_NOT_PASSED": STATUS_ERROR,
        }

    async def init(
        self,
        distros,
        timeout,
        reserve_duration,
        strategy=STRATEGY_ABORT,
        max_retry=1,
    ):
        """Initialize provider with data from Beaker configuration."""
        logger.info(f"{self.dsp_name} Initializing provider")
        self.strategy = strategy
        self.max_retry = max_retry
        self.distros = distros
        self.timeout = timeout
        self.reserve_duration = reserve_duration
        login_start = datetime.now()
        default_config = os.path.expanduser(
            os.environ.get("BEAKER_CONF", "/etc/beaker/client.conf")  # TODO use provc
        )  # get the beaker config for initialization of hub
        self.conf.load_from_file(default_config)
        try:
            self.hub = HubProxy(logger=logger, conf=self.conf)
        except MissingCredentialsError as kinit_err:
            raise NotAuthenticatedError(
                f"{self.dsp_name} needs Kerberos ticket to authenticate to BeakerHub. "
                "Run 'kinit $USER' command to obtain Kerberos credentials."
            ) from kinit_err
        except GSSError as hub_err:
            raise NotAuthenticatedError(
                f"{self.dsp_name} Unable to Create session: {hub_err}"
            ) from hub_err

        login_end = datetime.now()
        login_duration = login_end - login_start
        logger.info(f"{self.dsp_name} Init duration {login_duration}")

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        for req in reqs:
            req_dstr = req.get("distro")
            if not req.get("meta_distro") and req_dstr not in self.distros:
                raise ValidationError(
                    f"{self.dsp_name} provider does not support "
                    f"'{req_dstr}' distro in provisioning config"
                )
        return

    async def prepare_provisioning(self, reqs):
        """Prepare provisioning."""
        return bool(reqs)

    async def can_provision(self, hosts):
        """Check that hosts can be provisioned."""
        return True

    async def utilization(self):
        """Check percentage utilization of given provider."""
        return 0

    def _req_to_bkr_job(self, req):  # pylint: disable=too-many-locals
        """Transform requirement to beaker job xml."""
        specs = deepcopy(req)  # work with own copy, do not modify the input

        # Create recipe with the specifications
        recipe = BeakerRecipe(**specs)
        recipe.addBaseRequires(**specs)

        # Specify the architecture
        arch_node = xml_doc().createElement("distro_arch")
        arch_node.setAttribute("op", "=")
        arch_node.setAttribute("value", specs["arch"])
        recipe.addDistroRequires(arch_node)

        host_requires = specs.get("hostRequires")
        if host_requires:  # suppose to be dict like {"or": [dict()], "and": [dict()]}
            for operand, operand_value in host_requires.items():
                if operand.startswith("_"):
                    recipe.node.getElementsByTagName("hostRequires")[0].setAttribute(
                        operand[1:],
                        operand_value,
                    )
                    continue
                # known operands are ["and", "or"]
                req_node = xml_doc().createElement(operand)
                for dct in operand_value:
                    req_node = add_dict_to_node(req_node, dct)

                recipe.node.getElementsByTagName("hostRequires")[0].appendChild(
                    req_node
                )

        # Specify the custom xml distro_tag node with values from provisioning config
        distro_tags = specs.get("distro_tags")
        if distro_tags:
            for tag in distro_tags:
                tag_node = xml_doc().createElement("distro_tag")
                tag_node.setAttribute("op", "=")
                tag_node.setAttribute("value", tag)
                recipe.addDistroRequires(tag_node)

        # Add ReserveSys element to reserve system after provisioning
        recipe.addReservesys(duration=str(self.reserve_duration))

        for task in specs["tasks"]:
            recipe.addTask(task=task["name"], role=task["role"])

        # Create RecipeSet and add our Recipe to it.
        recipe_set = BeakerRecipeSet(**specs)
        recipe_set.addRecipe(recipe)

        # Create job instance and inject created RecipeSet to it
        job = BeakerJob(**specs)
        job.addRecipeSet(recipe_set)

        return job

    async def create_server(self, req):
        """Issue creation of a server.

        req - dict of server requirements

        The req object can contain following additional attributes:
        * 'name':       name for the VM
        * 'distro':     beaker distribution to use
        * 'arch':       architecture to request from beaker
        * 'variant':    variant of the system

        Returns:
            A tuple containing, respectively, a string (<created beaker job id>)
            and a dict (<requirements for VM>)
            :rtype: (str, dict)
        """
        logger.info(f"{self.dsp_name} [{req.get('name')}] Creating server")

        job = self._req_to_bkr_job(req)  # Generate the job
        try:
            job_id = self.hub.jobs.upload(job.toxml())  # schedule beaker job
        except Fault as bkr_fault:
            # use the name as id for the logging purposes
            req["host_id"] = req.get("name")
            raise ProvisioningError(
                parse_bkr_exc_str(bkr_fault),
                req,
            ) from bkr_fault

        return (job_id, req)

    def prov_result_to_host_data(self, prov_result, req):
        """Transform provisioning result to needed host data."""
        try:
            ip_address = socket.gethostbyname(prov_result["system"])
        except (TypeError, socket.gaierror):
            ip_address = None

        result = {
            "id": prov_result["JobID"],
            "name": prov_result.get("mrack_req").get("name"),
            "addresses": [ip_address],
            "status": prov_result["status"],
            "fault": None,
            "os": prov_result.get("mrack_req").get("os"),
            "group": prov_result.get("mrack_req").get("group"),
        }

        if prov_result["result"] != "Pass":
            result.update(
                {
                    "fault": prov_result["result"],
                    "status": "MRACK_RESULT_NOT_PASSED",
                }
            )

        return result

    def _get_recipe_info(self, beaker_id, log_msg_start):
        """Get info about the recipe for beaker job id."""
        bkr_job_xml = self.hub.taskactions.to_xml(beaker_id).encode("utf8")

        resources = []
        for recipe in eTree.fromstring(bkr_job_xml).iter("recipe"):
            resources.append(
                {
                    "system": recipe.get("system"),
                    "status": recipe.get("status"),
                    "result": recipe.get("result"),
                    "rid": recipe.get("id"),
                    "id": recipe.get("job_id"),
                }
            )

        logger.debug(
            f"{log_msg_start} has status:{resources[0]['status']}, "
            f"result:{resources[0]['result']}, waiting another {self.poll_sleep:.1f}s"
        )

        return resources[0]

    async def wait_till_provisioned(self, resource):
        """Wait for Beaker provisioning result."""
        beaker_id, req = resource
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"
        bkr_res = {}
        prev_status = ""
        job_url = ""
        hub_url = self.hub._hub_url  # pylint: disable=protected-access

        # let us use timeout variable which is in minutes to define
        # maximum time to wait for beaker recipe to provide VM
        timeout_time = datetime.now() + timedelta(minutes=self.timeout)

        while datetime.now() < timeout_time:
            prev_bkr_res = bkr_res
            try:
                bkr_res = self._get_recipe_info(beaker_id, log_msg_start=log_msg_start)
            except TimeoutError as timeout:
                logger.warning(
                    f"{log_msg_start} Can not connect to {hub_url}: {timeout}"
                )
                logger.debug(
                    f"{log_msg_start} Using previous result "
                    f"and retrying in {self.poll_sleep:.1f}s"
                )
                bkr_res = prev_bkr_res

            status = bkr_res.get("status", "")
            job_url = f"{hub_url}/jobs/{bkr_res.get('id', None)}"

            status_changed = prev_status != status
            if status_changed:
                logger.info(
                    f"{log_msg_start} Job {job_url} "
                    f"has changed status ({prev_status} -> {status})"
                )
                prev_status = status

            # if we have problem contacting hub from beginning: status will not change
            if self.status_map.get(status) == STATUS_PROVISIONING or not status_changed:
                await asyncio.sleep(self.poll_sleep)
            elif self.status_map.get(status) == STATUS_ACTIVE:
                break
            elif self.status_map.get(status) in [STATUS_ERROR, STATUS_DELETED]:
                logger.warning(
                    f"{log_msg_start} Job {job_url} has errored with status "
                    f"{status} and result {bkr_res['result']}"
                )
                bkr_res.update({"result": f"Job {job_url} failed to provision"})
                break
            else:
                logger.error(
                    f"{log_msg_start} Job {job_url} has switched to unexpected "
                    f"status {status} with result {bkr_res['result']}"
                )
                bkr_res.update({"result": f"Job {job_url} failed to provision"})
                break

        else:
            # In this case we failed to provision host in time:
            # we need to create failed host object for mrack
            # to delete the resource by cancelling the beaker job.
            logger.error(
                f"{log_msg_start} Job {job_url} failed to provide bkr_res in"
                f" the timeout of {self.timeout} minutes"
            )
            bkr_res.update(
                {
                    "status": "MRACK_REACHED_TIMEOUT",
                    "result": f"Job {job_url} reached timeout",
                }
            )

        bkr_res.update(
            {
                "JobID": beaker_id,
                "mrack_req": req,
            }
        )
        return bkr_res, req

    async def delete_host(self, host_id, host_name):
        """Delete provisioned hosts based on input from provision_hosts."""
        # host_id should start with 'J:' this way we know job has been scheduled
        # and proper response from beaker hub has beed returned.
        # Other way (In case of hub error or invalid host definition)
        # the provider uses hostname from metadata of the VM which has failed
        # to validate the requirements for the provider
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        if not host_id.startswith("J:"):
            logger.warning(
                f"{log_msg_start} Job for host '{host_id}' does not exist yet"
            )
            return True

        logger.info(
            f"{log_msg_start} Deleting host by cancelling Job "
            f"{self.hub._hub_url}"  # pylint: disable=protected-access
            f"/jobs/{host_id.split(':')[1]}"
        )
        return self.hub.taskactions.stop(
            host_id, "cancel", "Job has been stopped by mrack."
        )

    def to_host(self, provisioning_result, req, username="root"):
        """Transform provisioning result into Host object."""
        return super().to_host(provisioning_result, req, username)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Podman transformer module."""

from mrack.providers.provider import STRATEGY_ABORT
from mrack.transformers.transformer import DEFAULT_ATTEMPTS, Transformer
from mrack.utils import get_host_from_metadata

CONFIG_KEY = "podman"


class PodmanTransformer(Transformer):
    """Podman transformer."""

    _config_key = CONFIG_KEY
    _required_config_attrs = [
        "images",
        "pubkey",
        "default_network",
        "podman_options",
    ]

    async def init_provider(self):
        """Initialize associate provider and transformer display name."""
        self.dsp_name = "Podman"

        await self._provider.init(
            container_images=self.config["images"].values(),
            ssh_key=self.config["pubkey"],
            default_network=self.config["default_network"],
            container_options=self.config["podman_options"],
            extra_commands=self.config.get("extra_commands", []),
            strategy=self.config.get("strategy", STRATEGY_ABORT),
            max_retry=self.config.get("max_retry", DEFAULT_ATTEMPTS),
        )

    def create_host_requirement(self, host):
        """Create single input for podman provisioner."""
        _host, domain = get_host_from_metadata(self._metadata, host["name"])
        return {
            "name": host["name"],
            "image": self._get_image(host),
            "os": host["os"],
            "group": host["group"],
            "hostname": host["name"],
            "domain": domain["name"],
        }



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Static provider."""

import logging
from copy import deepcopy

from mrack.errors import ValidationError
from mrack.host import STATUS_ACTIVE
from mrack.providers.provider import STRATEGY_ABORT, Provider

PROVISIONER_KEY = "static"

logger = logging.getLogger(__name__)


class StaticProvider(Provider):
    """
    Static Provider.

    Doesn't provision anything. Only simulates provisioning so that
    we can track the server in output modules and other actions.
    """

    def __init__(self):
        """Object initialization."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "Static"
        self.max_retry = 1  # for retry strategy
        self.strategy = STRATEGY_ABORT
        self.status_map = {STATUS_ACTIVE: STATUS_ACTIVE}

    async def prepare_provisioning(self, reqs):
        """Prepare provisioning."""
        return bool(reqs)

    async def create_server(self, req):
        """Request and create resource on selected provider."""
        # we shall return same tuple as with other providers
        return (req, deepcopy(req))

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        for req in reqs:
            if "name" not in req:
                raise ValidationError("Name not found")
            if "ip" not in req:
                raise ValidationError("IP address (ip) not found")
        return bool(reqs)

    async def can_provision(self, hosts):
        """Behave as that we can."""
        return True

    async def utilization(self):
        """Check percentage utilization of given provider."""
        return 0

    async def provision_hosts(self, reqs):
        """Fake provision - behave as if it was provisioned."""
        await self.validate_hosts(reqs)
        hosts = [self.to_host(req, req) for req in reqs]
        return hosts

    async def wait_till_provisioned(self, resource):
        """Wait till resource is provisioned."""
        return resource

    def prov_result_to_host_data(self, prov_result, req):
        """Transform provisioning result to needed host data."""
        result = {}

        result["id"] = prov_result.get("name")
        result["name"] = req.get("name")
        result["addresses"] = [prov_result.get("ip")]
        result["fault"] = {}
        result["status"] = STATUS_ACTIVE
        result["os"] = prov_result.get("os")  # with static it stays os
        result["group"] = prov_result.get("group")  # with static it stays os

        return result

    async def delete_host(self, host_id, host_name):
        """Fake delete - pass but don't do anything."""
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        logger.info(
            f"{log_msg_start} Host with ID {host_id} is static"
            " - not managed by mrack, ignoring removal"
        )
        return bool(host_id)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SSH action module."""

import logging

from mrack.actions.action import Action
from mrack.context import global_context
from mrack.errors import ApplicationError
from mrack.host import STATUS_ACTIVE
from mrack.utils import get_username_pass_and_ssh_key
from mrack.utils import ssh_to_host as utils_ssh_to_host

try:
    from mrack.providers.utils.podman import Podman
except ModuleNotFoundError:
    pass  # Ignore the module import error if plugin not available

logger = logging.getLogger(__name__)


class SSH(Action):
    """SSH action.

    SSH all still active provisioned host. Save the state to DB.
    """

    def pick_host_interactively(self, hosts):
        """Print list of host and let user to choose one."""
        # With only 1 host available, we can pick it and save time
        if len(hosts) == 1:
            return hosts[0].name

        for idx, host in enumerate(hosts):
            logger.info(f"{idx}: {host}")

        while True:
            try:
                host_idx = int(input("Enter a number of host to ssh into: "))
            except ValueError:
                logger.info("Entered value is not a number.")
                continue
            if host_idx < 0 or host_idx >= len(hosts):
                logger.info("Entered value is not a number of any host.")
                continue
            break

        return hosts[host_idx].name

    def find_host(self, hostname):
        """Find active host based on hostname or interactively."""
        hosts = self._db_driver.hosts
        active_hosts = [h for h in hosts.values() if h.status == STATUS_ACTIVE]

        if not hosts:
            raise ApplicationError("No hosts provisioned.")

        if not active_hosts:
            raise ApplicationError("No active host available.")

        if not hostname:
            hostname = self.pick_host_interactively(active_hosts)

        host = hosts.get(hostname)
        if not host:
            raise ApplicationError("Specified host does not exist.")

        if host.status != STATUS_ACTIVE:
            raise ApplicationError("Selected host is not active.")

        return host

    def ssh_to_host(self, host):
        """SSH to the selected host."""
        username, password, ssh_key = get_username_pass_and_ssh_key(
            host, global_context
        )
        return utils_ssh_to_host(
            host,
            username=username,
            password=password,
            ssh_key=ssh_key,
            interactive=True,
        )

    def is_container_env(self, host):
        """Check if host is using local container technology."""
        return host.provider.name in ["docker", "podman"]

    def attach_interactive_container(self, host):
        """Simulate SSH by attaching an interactive session to a container."""
        if host.provider.name == "podman":
            podman = Podman()
            podman.interactive(host.host_id)
        else:
            raise NotImplementedError("Docker is not yet supported.")

    def ssh(self, hostname):
        """Execute the SSH action."""
        host = self.find_host(hostname)
        if self.is_container_env(host):
            self.attach_interactive_container(host)
        else:
            self.ssh_to_host(host)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provisioning providers.

Providers provisions resources in Clouds or other means based on provisioning
requirements (e.g. host requirements)
"""

from mrack.errors import ProviderNotExists


class Registry:
    """Provider registry."""

    def __init__(self):
        """Initialize Registry."""
        self._provider_cls = {}
        self._providers = {}

    def register(self, name, provider_cls):
        """Register new provider."""
        self._provider_cls[name] = provider_cls

    def get(self, name):
        """Get a provider object by name.

        Instantiate the provider if not done yet.

        Raises: ProviderNotExists
        """
        prov_cls = self._provider_cls.get(name)
        if not prov_cls:
            raise ProviderNotExists(f"Provider '{name}' doesn't exist")
        provider = self._providers.get(name)
        if not provider:
            provider = prov_cls()
            self._providers[name] = provider
        return provider

    @property
    def names(self):
        """Get all registered provider names."""
        return self._provider_cls.keys()


providers = Registry()
# registersing in run.py init_providers()



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic Transformer."""

import logging
import os
import sys
import typing

from mrack.context import global_context
from mrack.errors import ConfigError, MetadataError, ValidationError
from mrack.providers import providers
from mrack.utils import (
    find_value_in_config_hierarchy,
    get_config_value,
    object2json,
    validate_dict_attrs,
)

if sys.version_info >= (3, 9):
    from functools import cache
else:
    from functools import lru_cache

    # Add cache decorator for the releases using python <= 3.9, e.g. c8s with py3.6
    def cache(user_function):
        """
        Add simple lightweight unbounded cache.

        Sometimes called "memoize".
        """
        return lru_cache(maxsize=None)(user_function)


DEFAULT_ATTEMPTS = 1

logger = logging.getLogger(__name__)

IMPORT_ERR = "gssapi not imported, VM owner info fetch from krb will fail"
try:
    import gssapi
except ModuleNotFoundError as import_err:
    logger.debug(IMPORT_ERR, import_err.name)


@cache
def cached_get_vm_owner():
    """Return ownership info for host."""
    owner = os.getenv("MRACK_VM_OWNER")
    logger.debug(f"MRACK_VM_OWNER: {owner}")
    if not owner and gssapi:
        try:
            creds = gssapi.Credentials(usage="initiate")
            logger.debug(f"GSSAPI credentials: {creds}")
            owner = str(creds.name).split("@", maxsplit=1)[0]
        except gssapi.exceptions.MissingCredentialsError:
            logger.debug("No kerberos credentials discovered.")
        except gssapi.exceptions.GSSError as exc:
            logger.exception(exc)
    return owner


@cache
def cached_get_vm_lifetime():
    """Return lifetime info for host."""
    return os.getenv("MRACK_VM_LIFETIME")


class Transformer:
    """Base class for transformers."""

    _required_host_attrs: typing.List[str] = ["name", "os", "group"]
    _required_config_attrs: typing.List[str] = []
    _config_key = ""

    async def init(self, cfg, metadata):
        """Initialize transformer."""
        self.dsp_name = "Transformer"
        self._hosts = []
        self._config = cfg
        self._metadata = metadata
        if self._config_key:
            self.validate_config()

        self._provider = providers.get(self._config_key)
        await self.init_provider()

    async def init_provider(self):
        """Initialize associated provider."""
        pass

    @property
    def config(self):
        """Get transformer/provider configuration from provisioning configuration."""
        key = self._config_key
        try:
            return self._config[key]
        except KeyError as key_err:
            error = f"No '{key}' entry in provisioning configuration"
            raise ConfigError(error) from key_err

    @property
    def hosts(self):
        """Get all host inputs this transformer is handling."""
        return self._hosts

    def add_host(self, host):
        """Add host input."""
        self._hosts.append(host)

    def _find_value(self, host, attr, dict_name, key, default=None):
        """Find value in config hierarchy."""
        value = find_value_in_config_hierarchy(
            self._config, self._config_key, None, host, attr, dict_name, key, default
        )
        return value

    def _get_image(self, host):
        """
        Get image name by OS name from provisioning config.

        The preferred solution is to create mappings between operating system
        and image present in the provisioning config - this only allows us
        to change the providers without changing the input metadata.

        host - dictionary with server requirements

        The host dictionary can contain following attributes:
        * 'os': operating system to find mapping to image from provisioning config
        * 'image': image to use (optional)

        Returns:
            An image to be used, if mapping is not found it defaults to the host['os']
            :rtype: str
        """
        operating_system = host["os"]
        image = self._find_value(
            host, "image", "images", operating_system, default=operating_system
        )

        logger.debug(
            f"{self.dsp_name} [{host.get('name')}] Using image "
            f"{image} for {operating_system}"
        )
        return image

    def _get_flavor(self, host):
        """
        Get VM flavor from host metadata definition or based on host group.

        Get the flavor from 'size' definition in host metadata.
        If 'size' is not specified, get the flavor based on host groups
        which are defined in provisioning config.
        """
        log_msg_start = f"{self.dsp_name} [{host.get('name')}]"

        if host.get("size"):  # allow to override size by pointing to group in metadata
            flavor = get_config_value(self.config["flavors"], host["size"])
        else:  # default action based on host group define the flavor of instance
            flavor = get_config_value(self.config["flavors"], host["group"])

        logger.debug(f"{log_msg_start} Loaded flavor '{flavor}'")
        return flavor

    def validate_config(self):
        """Validate provisioning configuration for this transformer/provider."""
        validate_dict_attrs(self.config, self._required_config_attrs, "config")

    def validate_host(self, host):
        """Validate host input that it contains everything needed by provider."""
        log_msg_start = f"{self.dsp_name} [{host.get('name')}]"
        # attribute check
        validate_dict_attrs(host, self._required_host_attrs, "host")
        # provider check
        provider = host.get("provider")

        if provider and provider not in providers.names:
            raise MetadataError(
                f"{log_msg_start} Error: Invalid host provider: {provider}"
            )

    def create_host_requirement(self, host):
        """Create single input for provisioner."""
        raise NotImplementedError()

    def create_host_requirements(self):
        """Create inputs for all host for provisioner."""
        reqs = [self.create_host_requirement(host) for host in self.hosts]
        logger.info(
            f"{self.dsp_name} Created {len(reqs)} requirement(s): {object2json(reqs)}"
        )
        return reqs

    def get_vm_owner(self):
        """Return ownership info for host."""
        return cached_get_vm_owner()

    def get_vm_lifetime(self):
        """Return lifetime info for host."""
        return cached_get_vm_lifetime()

    def validate_ownership_and_lifetime(self, host):
        """Validate if ownership and lifetime exists for host."""
        log_msg_start = f"{self.dsp_name} [{host.get('name')}]"
        require_owner = global_context.CONFIG.require_owner()
        owner = self.get_vm_owner()
        if require_owner and not owner:
            raise ValidationError(
                f"{log_msg_start} VM ownership information is required. "
                "Run kinit and try provisioning again"
            )

        lifetime = self.get_vm_lifetime()
        if lifetime:
            try:
                val = int(lifetime)
            except ValueError as value_err:
                error = f"{log_msg_start} VM lifetime must be integer"
                raise ValidationError(error) from value_err
            if val < 0:
                raise ValidationError(
                    f"{log_msg_start} VM lifetime must be greater than 0"
                )
            if not owner:
                raise ValidationError(
                    f"{log_msg_start} Owner is required when vm lifetime is set"
                )

    def update_metadata_for_owner_lifetime(self, req):
        """Update VM metadata for owner and lifetime."""
        owner = self.get_vm_owner()
        lifetime = self.get_vm_lifetime()
        metadata = req.get("metadata", {})
        if owner:
            metadata["owner"] = owner
        if lifetime:
            metadata["lifetime"] = lifetime
        if owner or lifetime:
            req["metadata"] = metadata
        return req



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to work with provisioning config.

Provisioning config is general project configuration shared between multiple jobs
"""

import logging
import os
from configparser import ConfigParser, NoOptionError, ParsingError

from mrack.errors import ConfigError
from mrack.utils import value_to_bool

logger = logging.getLogger(__name__)


class ProvisioningConfig:
    """Represents loaded provisioning configuration."""

    def __init__(self, data):
        """Initialize provisioning configuration."""
        self._raw = data

    def raw(self):
        """Get raw configuration."""
        return self._raw()

    def __getitem__(self, key):
        """Get item from raw representation."""
        return self._raw[key]

    def get(self, key, default=None):
        """Get method as in dict for raw config."""
        return self._raw.get(key, default)


class MrackConfig:
    """Configuration for mrack itself."""

    def __init__(self, user_provided_path=None):
        """Init Mrack Config."""
        self._user_provided_path = user_provided_path
        self.config_name = "mrack.conf"
        self.section = "mrack"
        self._parser = ConfigParser()

    def get_config_paths(self):
        """Get possible mrack configuration file paths ordered by priority."""
        cwd = os.path.abspath(".")
        home = os.path.expanduser("~")
        paths = [
            f"{cwd}/{self.config_name}",
            f"{home}/.mrack/{self.config_name}",
            f"/etc/mrack/{self.config_name}",
        ]
        if self._user_provided_path:
            paths.insert(0, self._user_provided_path)
        return paths

    def load(self):
        """Load mrack configuration, first config file wins, rest is ignored."""
        cfg_paths = self.get_config_paths()
        chosen = None
        for cfg_path in cfg_paths:
            if os.path.exists(cfg_path):
                chosen = cfg_path
                break

        if not chosen:
            logger.debug("No config file found.")
            self._parser.read_dict({"mrack": {}})  # empty mrack section
            return

        try:
            logger.debug(f"Loading config file: {chosen}.")
            self._parser.read(chosen)
        except ParsingError as parse_err:
            raise ConfigError(
                f"Invalid syntax in configuration file {chosen}."
            ) from parse_err

        if not self._parser.has_section(self.section):
            raise ConfigError(
                f"Configuration file {chosen} doesn't have "
                f"required section [{self.section}]."
            )

        return

    def __getitem__(self, key):
        """Get configuration value."""
        try:
            value = self._parser.get(self.section, key)
            return value
        except NoOptionError as no_opt_err:
            raise KeyError(
                f"Configuration file doesn't have key: {key}"
            ) from no_opt_err

    def get(self, key, default=None):
        """Get configuration value."""
        try:
            value = self._parser.get(self.section, key)
            logger.debug(f"Config key: {key}, value: {value}")
            return value
        except NoOptionError:
            return default

    def provisioning_config_path(self, default=None):
        """Return configured provisioning config path."""
        return self.get("provisioning-config", default)

    def db_path(self, default=None):
        """Return configured Mrack database path."""
        return self.get("mrackdb", default)

    def metadata_path(self, default=None):
        """Return configured job metadata path."""
        return self.get("metadata", default)

    def ansible_inventory_path(self, default=None):
        """Return configured path to Ansible inventory output."""
        return self.get("ansible-inventory", default)

    def pytest_multihost_path(self, default=None):
        """Return configured path to pytest-multihost configuration output."""
        return self.get("pytest-multihost", default)

    def pytest_mh_path(self, default=None):
        """Return configured path to pytest-mh configuration output."""
        return self.get("pytest-mh", default)

    def require_owner(self, default=False):
        """Return value of require-owner."""
        return value_to_bool(self.get("require-owner", default))

    @property
    def delta_sleep(self):
        """Return value of `delta-sleep` value from config to randomize sleep window."""
        return int(self.get("delta-sleep", default=15))

    @property
    def max_utilization(self):
        """Return value `max-utilization` from mrack config file."""
        return int(self.get("max-utilization", default=90))

    @property
    def usable_network_threshold(self):
        """Return maximum acceptable network utilization of provider."""
        return int(self.get("usable-network-threshold", default=95))

    @property
    def network_spread(self):
        """Return `network-spread` setting value from mrack config.

        Possible values are:
        - no: disable network spreading feature
        - allow: allow network spreading feature (default)
        - force: always use network spreading feature
        """
        return self.get("network-spread", default="allow").lower()



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AWS transformer module."""

from mrack.providers.provider import STRATEGY_ABORT
from mrack.transformers.transformer import DEFAULT_ATTEMPTS, Transformer

CONFIG_KEY = "aws"


class AWSTransformer(Transformer):
    """AWS transformer."""

    _config_key = CONFIG_KEY
    _required_config_attrs = [
        "flavors",
        "images",
        "credentials_file",
        "keypair",
        "security_groups",
        "profile",
        "instance_tags",
    ]

    async def init_provider(self):
        """Initialize associate provider and transformer display name."""
        self.dsp_name = "AWS"
        await self._provider.init(
            ssh_key=self.config["keypair"],
            instance_tags=self.config["instance_tags"],
            strategy=self.config.get("strategy", STRATEGY_ABORT),
            max_retry=self.config.get("max_retry", DEFAULT_ATTEMPTS),
        )

    def _get_security_groups(self):
        """
        Get Security Group IDs.

        Works with both single `subnet_id` definition as well as with newer
        `subnet_ids` which allows to defined multiple.
        """
        sc_ids = set(self.config.get("security_groups", []))
        sc_id = self.config.get("security_group")
        if sc_id:
            sc_ids.add(sc_id)
        return list(sc_ids)

    def _find_subnet_ids(self, host):
        """Get subnet id/s from config.

        Note: 'subnet_ids' has preference against legacy way 'subnet_id'.

        Returns: List of found subnet ids
        """
        subnet_ids = self._find_value(
            host.get(CONFIG_KEY, {}),
            "subnet_ids",
            "subnet_ids",
            host["os"],
            default=[],
        )
        subnet_id = self._find_value(
            host.get(CONFIG_KEY, {}),
            "subnet_id",
            None,
            None,
            default="",
        )

        if "subnet_ids" in host.get(CONFIG_KEY, {}):
            return subnet_ids
        if "subnet_id" in host.get(CONFIG_KEY, {}):
            return [subnet_id]
        if subnet_ids:
            return subnet_ids
        if subnet_id:
            return [subnet_id]

        return []

    def validate_host(self, host):
        """Validate host input that it contains what AWS needs."""
        super().validate_host(host)
        self.validate_ownership_and_lifetime(host)

    def create_host_requirement(self, host):
        """Create single input for AWS provisioner."""
        del_vol = self._find_value(
            host, "delete_volume_on_termination", None, None, True
        )
        req = {
            "name": host["name"],
            "os": host["os"],
            "group": host["group"],
            "flavor": self._get_flavor(host),
            "image": self._get_image(host),
            "security_group_ids": self._get_security_groups(),
            "spot": self._find_value(host, "spot", None, None),
            "delete_volume_on_termination": del_vol,
            "subnet_ids": self._find_subnet_ids(host),
        }

        req = self.update_metadata_for_owner_lifetime(req)
        return req



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provisioning errors."""


class MrackError(Exception):
    """Base mrack error."""

    pass


class ConfigError(MrackError):
    """Error in configuration."""

    pass


class ApplicationError(MrackError):
    """General application error."""

    pass


class MetadataError(MrackError):
    """Error in job metadata."""

    pass


class ProviderNotExists(ConfigError):
    """Request provider does not exist."""

    def __init__(self, name):
        """Init the error with provider name."""
        self._provider_name = name


class ProvisioningConfigError(ConfigError):
    """Error in provisioning configuration."""

    pass


class JobConfigError(ConfigError):
    """Error in job configuration."""

    pass


class ValidationError(MrackError):
    """Error found in validation of values."""

    pass


class ProviderError(MrackError):
    """General provider error."""

    pass


class ServerNotFoundError(ProviderError):
    """Provider doesn't know the specified server."""

    pass


class ProvisioningError(ProviderError):
    """Error happened during provisioning of resources."""

    pass


class NotAuthenticatedError(ProviderError):
    """Provided is not authenticated."""

    pass



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""pytest-multihost configuration file output module."""

import logging
import os
from copy import deepcopy

from mrack.outputs.utils import get_external_id
from mrack.utils import get_password, get_username, is_windows_host, save_yaml

DEFAULT_MHCFG_PATH = "pytest-multihost.yaml"

logger = logging.getLogger(__name__)


class PytestMultihostOutput:
    """
    Create a configuration file for python-pytest-multihost.

    From job metadata and provisioned information (DB).

    Structure of multihost config file is almost the same as our topology in job
    metadata definition.
    """

    def __init__(self, config, db, metadata, path=None):  # pylint: disable=invalid-name
        """Init the output module."""
        self._config = config
        self._db = db
        self._metadata = metadata
        self._path = path or DEFAULT_MHCFG_PATH

    def create_multihost_config(self):  # pylint: disable=too-many-branches
        """
        Create configuration file for python-pytest-multihost.

        Return in form of dict.
        """
        mhcfg = deepcopy(self._metadata)

        if "phases" in mhcfg:
            del mhcfg["phases"]

        if "config" in mhcfg:
            del mhcfg["config"]

        # topology config in metadata contains more values (os, group) than the ones
        # needed by pytest multihost. They need to be removed in order to work.
        host_allowed_attrs = [
            "name",
            "role",
            "hostname",
            "shortname",
            "external_hostname",
            "ip",
            "domain",
            "username",
            "password",
            "host_type",
        ]

        # Use values from provisioning-config mhcfg section - as default values
        # options defined in metadata takes precedence over options in
        # provisioning-config as it is "closer" to user.
        mhcfg = self._config.get("mhcfg", {}) | mhcfg

        for domain in mhcfg["domains"]:
            for host in domain["hosts"]:
                provisioned_host = self._db.hosts.get(host["name"])
                if not provisioned_host:
                    logger.error(f"Host {host['name']} not found in the database.")
                    continue

                password = get_password(provisioned_host, host, self._config)
                # pytest-multihost doesn't support different ssh_keys per host

                if password:
                    host["password"] = password

                host["ip"] = provisioned_host.ip_addr

                # Using IP as backup for external host name as pytest-multihost is using
                # external_hostname as the host to use in ssh command.
                # If it is not available it uses hostname, but we assume here that
                # hostname is internal and thus not resolvable. IP should be resolvable.
                host["external_hostname"] = get_external_id(
                    provisioned_host, host, self._config
                )

                if is_windows_host(host):
                    # Set username for Windows, as default for multihost is often 'root'
                    # which doesn't usually work there. But do not set it for other OSes
                    # due to:
                    #  7bb230e170ac0a2373a2316ef23a26bfcb681ad9
                    # TODO: come up with a configurable mechanism which is not based on
                    # so many assumptions.
                    username = get_username(provisioned_host, host, self._config)
                    if username:
                        host["username"] = username

                group = host.get("group")
                if not group:
                    groups = host.get("groups")
                    if isinstance(groups, list) and len(groups):
                        group = groups[0]
                    else:
                        group = ""  # group is required, but not part of output

                if group == "ad_root":
                    mhcfg["ad_top_domain"] = domain["name"]
                    mhcfg["ad_hostname"] = host["name"]
                    mhcfg["ad_ip"] = host["ip"]
                elif group == "ad_subdomain":
                    mhcfg["ad_sub_domain"] = domain["name"]
                    mhcfg["ad_sub_hostname"] = host["name"]
                    mhcfg["ad_sub_ip"] = host["ip"]

                host_custom_attrs = deepcopy(host.get("pytest_multihost", {}))

                rm_keys = [key for key in host.keys() if key not in host_allowed_attrs]
                for key in rm_keys:
                    del host[key]

                if host_custom_attrs:
                    host.update(host_custom_attrs)

            domain_rm_keys = [
                key for key in domain.keys() if key not in ["name", "type", "hosts"]
            ]
            for key in domain_rm_keys:
                del domain[key]

        # ssh_key_filename must be absolute as it can be used from a different
        # working directory, e.g. running tests from git repo
        ssh_key_filename = mhcfg.get("ssh_key_filename")

        # Update with SSH key path only if not already defined in mhcfg
        if ssh_key_filename and not self._metadata.get("ssh_key_filename"):
            mhcfg["ssh_key_filename"] = os.path.abspath(ssh_key_filename)

        return mhcfg

    def create_output(self):
        """Create the target output file."""
        mhcfg = self.create_multihost_config()
        save_yaml(self._path, mhcfg)
        if self._path:
            logger.info(f"Created: {self._path}")
        return mhcfg



"""mrack library."""

import logging

logger = logging.getLogger(__name__)
# Default level used by file_handler
logger.setLevel(logging.DEBUG)

file_formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
console_formatter = logging.Formatter("%(message)s")

file_handler = logging.FileHandler("mrack.log")
file_handler.setFormatter(file_formatter)

console_handler = logging.StreamHandler()
console_handler.setFormatter(console_formatter)

# Default level for console can be overridden by passing --debug to the executable
console_handler.setLevel(logging.INFO)

logger.addHandler(file_handler)
logger.addHandler(console_handler)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Up action module."""

import asyncio
import logging

from mrack.actions.action import Action
from mrack.errors import MetadataError, MrackError, ProvisioningError
from mrack.providers import providers
from mrack.utils import validate_dict_attrs

ERR_MSG_INDEX = 0
PROVIDER_NAME_INDEX = 1

logger = logging.getLogger(__name__)


class Up(Action):
    """
    Up action.

    UP action provision all hosts defined in domain object of provided job
    metadata.

    The provisioning information is created by transformers from metadata
    and provisioning configuration (config).

    Results of provisioning is saved by 'db_driver'.

    Output objects generate the resulting output artifacts, e.g. Ansible
    inventory.
    """

    def __init__(self, config=None, metadata=None, db_driver=None):
        """Initialize the Up action."""
        super().__init__(config=config, metadata=metadata, db_driver=db_driver)
        self._required_domain_attrs = ["name", "hosts"]

    async def init(self, default_provider):
        """Initialize the Up action for the topology."""
        self.validate_topology()
        default_provider_name = self._config.get("provider", default_provider)
        default_provider_name = self._metadata.get("provider", default_provider_name)
        for domain in self._metadata["domains"]:
            for host in domain["hosts"]:
                provider_name = host.get("provider", default_provider_name)
                transformer = await self._get_transformer(provider_name)
                transformer.validate_host(host)
                transformer.add_host(host)

    def validate_topology(self):
        """Validate topology part of job metadata.

        Raises: MetadataError
        """
        if "domains" not in self._metadata:
            error = (
                "Error: job metadata file doesn't contain required "
                "`domains` definition"
            )
            raise MetadataError(error)
        for domain in self._metadata["domains"]:
            validate_dict_attrs(domain, self._required_domain_attrs, "domain")

    async def provision(self):
        """Execute the up action."""
        logger.info("Provisioning started")
        prov_aws = []
        for provider_name, transformer in self._transformers.items():
            reqs = transformer.create_host_requirements()
            provider = providers.get(provider_name)
            awaitable = provider.provision_hosts(reqs)
            prov_aws.append(awaitable)

        provisioning_results = await asyncio.gather(*prov_aws, return_exceptions=True)

        success_hosts = []
        failed_providers = []
        for results in provisioning_results:
            if isinstance(results, MrackError):
                # in this case some of any provider fails to provision
                # we need to cleanup all the remaining resources
                # even when provisioned successfully
                logger.error(
                    f"{results.args[PROVIDER_NAME_INDEX]} "
                    f"{results.args[ERR_MSG_INDEX]}"
                )
                failed_providers.append(results.args[PROVIDER_NAME_INDEX])
                continue

            if isinstance(results, Exception):
                logger.error("An unexpected exception occurred while provisioning")
                raise results

            success_hosts.extend(results)

        if failed_providers and success_hosts:
            # if there is successfully provisioned resource do a cleanup
            logger.info(
                "Issuing deletion of all successfully provisioned resources "
                " due to provisioning error"
            )
            cleanup = [h.delete() for h in success_hosts]
            cleanup_res = await asyncio.gather(*cleanup)
            if all(cleanup_res):
                logger.info(
                    "All successfully provisioned resources issued to be deleted"
                )
            else:
                logger.error(
                    "Failed to destroy some of provisioned resources, "
                    "manual destroy of resources might be needed"
                )

        if failed_providers:
            failed_prov = ", ".join(failed_providers)
            raise ProvisioningError(
                f"Provider(s) {failed_prov} failed to provision resources"
            )

        self._db_driver.add_hosts(success_hosts)
        logger.info("Provisioning done")
        return success_hosts



# Copyright 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for working with testcloud."""

# https://pagure.io/testcloud


import asyncio
import logging

from testcloud import config as tc_config
from testcloud import instance as tc_instance
from testcloud.exceptions import TestcloudPermissionsError
from testcloud.image import Image

from mrack.utils import exec_async_subprocess

logger = logging.getLogger(__name__)


class Testcloud:
    """Async wrapper supporting most basic testcloud calls."""

    def __init__(self, program="testcloud"):
        """Init the instance."""
        self.program = program

    async def _run_testcloud(self, args, raise_on_err=True):
        """Util method to execute testcloud process."""
        logger.info("Running: testcloud %s", " ".join(args))
        return await exec_async_subprocess(self.program, args, raise_on_err)

    async def list(self):
        """List instances of testcloud."""
        return await self._run_testcloud(["instance", "list"])

    async def clean(self):
        """Remove non-existing libvirt VMs from testcloud."""
        return await self._run_testcloud(["instance", "clean"])

    async def _instance_command(self, command, instance_name, additional_args=None):
        """Run instance command of test cloud."""
        args = ["instance", command, instance_name]
        if additional_args:
            args.extend(additional_args)
        return await self._run_testcloud(args)

    async def create(
        self,
        instance_name,
        image_url,
        **kwargs,
    ):
        """Start a new testcloud instance."""
        args = ["instance", "create", "--name", instance_name]

        if kwargs.get("ram"):
            args.extend(["--ram", kwargs.get("ram")])
        if kwargs.get("vcpus"):
            args.extend(["--vcpus", kwargs.get("vcpus")])
        if kwargs.get("disksize"):
            args.extend(["--disksize", kwargs.get("disksize")])
        if kwargs.get("timeout"):
            args.extend(["--timeout", kwargs["timeout"]])
        if kwargs.get("vnc"):
            args.extend(["--vnc"])
        if kwargs.get("no_graphics"):
            args.extend(["--no_graphics"])
        if kwargs.get("keep"):
            args.extend(["--keep"])
        if kwargs.get("ssh_path"):
            args.extend(["--ssh_path", kwargs.get("ssh_path")])

        args.append(image_url)

        return await self._run_testcloud(args)

    def info(self, instance_name):
        """Find instance information."""
        logger.debug(f"running info {instance_name}")
        instances = tc_instance.list_instances()
        config_data = tc_config.get_config()
        match = [i for i in instances if i["name"] == instance_name]
        if match:
            inst = match[0]
            logger.debug(inst)
            return {
                "name": inst["name"],
                "ip": inst["ip"],
                "port": inst["port"],
                "state": inst["state"],
                "password": config_data.PASSWORD,
            }
        return None

    async def start(self, instance_name):
        """Stop an instance."""
        return await self._instance_command("start", instance_name)

    async def stop(self, instance_name):
        """Stop an instance."""
        return await self._instance_command("stop", instance_name)

    async def destroy(self, instance_name):
        """Remove an instance."""
        return await self._instance_command("destroy", instance_name, ["-f"])

    async def reboot(self, instance_name):
        """Reboot an instance."""
        return await self._instance_command("reboot", instance_name)

    def _pull_image(self, image_url):
        """Pull image in testcloud image store."""
        tc_image = Image(image_url)
        try:
            tc_image.prepare()
            return True
        except TestcloudPermissionsError as error:
            logger.error(error)
            return False

    async def pull_image(self, image_url):
        """Pull image in testcloud image store."""
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._pull_image, image_url)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""mrack setup"""
# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open("requirements.txt", encoding="utf-8") as req:
    reqs = req.readlines()

with open("src/mrack/version.py", "r", encoding="utf-8") as fd:
    version = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    if version is None:
        sys.stderr.write("Could not parse the version string.\n")
        sys.exit(1)

    mrack_version = version.group(1)


MRACK_CONF = "mrack.conf"
PROV_CONF = "provisioning-config.yaml"

setup(
    name="mrack",
    version=mrack_version,
    description="Multicloud use-case based multihost async provisioner "
    "for CIs and testing during development",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Petr Vobornik",
    author_email="pvoborni@redhat.com",
    url="https://github.com/neoave/mrack",
    license="Apache License 2.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=reqs,
    scripts=["scripts/mrack"],
    package_data={
        "mrack": [f"data/{datafile}" for datafile in [MRACK_CONF, PROV_CONF]]
    },
    extras_require={
        "krb-owner": ["gssapi"],
    },
)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Etc-Hosts actions."""

import getpass
import logging

from mrack.actions.action import DBAction
from mrack.errors import ApplicationError, ConfigError
from mrack.host import STATUS_ACTIVE

logger = logging.getLogger(__name__)

MRACK_START = "# Managed by mrack - start\n"
MRACK_END = "# Managed by mrack - end\n"


class EtcHostsUpdater:
    """Object to manipulate with /etc/hosts file."""

    def __init__(self, path="/etc/hosts"):
        """Init the updater."""
        self.path = path
        with open(self.path, "r", encoding="utf-8") as etc_hosts:
            self.lines = etc_hosts.readlines()

        # Trigger validation
        self._locate_mrack_section()

    def _locate_mrack_section(self):
        """Locate mrack managed section."""
        start = -1
        end = -1
        mrack_lines = []
        errors = []
        for index, line in enumerate(self.lines):
            if MRACK_START in line:
                if start != -1:
                    errors.append("Multiple start marks.")
                start = index
            if MRACK_END in line:
                if end != -1:
                    errors.append("Multiple end marks.")
                end = index
            if start != -1 and end == -1 and index > start:
                mrack_lines.append([index, line])

        # Check errors:
        if start == -1 and end != -1:
            errors.append("Doesn't have start mark.")
        if start != -1 and end == -1:
            errors.append("Doesn't have end mark.")
        if end < start:
            errors.append("End mark before start mark.")

        if errors:
            error = "/etc/hosts has broken mrack managed section and needs fixing:\n"
            for err in errors:
                error += f"   {err}\n"
            raise ConfigError(error)

        return (start, end, mrack_lines)

    def add_hosts(self, hosts):
        """Add lines for hosts."""
        start, end, mrack_lines = self._locate_mrack_section()

        if start == -1:
            start = len(self.lines)
            self.lines.append(MRACK_START)
            end = start + 1
            self.lines.append(MRACK_END)

        for host in hosts:
            new_line = f"{host.ip_addr} {host.name}\n"
            for mrack_line in mrack_lines:
                idx, line = mrack_line
                # update: /etc/hosts has already a line for the host
                if host.name in line:
                    self.lines[idx] = new_line
                    mrack_line[1] = new_line
                    break
            else:
                mrack_lines.append([end, new_line])
                self.lines.insert(end, new_line)
                end += 1

    def clear(self):
        """Remove mrack managed lines."""
        start, end, _mrack_lines = self._locate_mrack_section()
        if start:
            del self.lines[start : end + 1]

    def save(self):
        """Save changes into /etc/hosts."""
        try:
            with open(self.path, "w", encoding="utf-8") as etc_file:
                etc_file.writelines(self.lines)
        except PermissionError as perm_err:
            name = getpass.getuser()
            err = (
                f"Error: Not enough permissions to write into file: {self.path}.\n\n"
                "Consider running mrack as root for this command e.g. via sudo "
                "or add yourself the necessary rights, e.g. by: \n"
                f"  $ sudo setfacl -m {name}:rw {self.path}"
            )
            raise ApplicationError(err) from perm_err


class EtcHostsUpdate(DBAction):
    """Add active vms to /etc/hosts file."""

    def __init__(self, db_driver=None):
        """Initialize the /etc/hosts update action."""
        super().__init__(db_driver=db_driver)
        self.updater = EtcHostsUpdater()

    def update(self):
        """Update /etc/hosts."""
        logger.info("Adding hosts to /etc/hosts file")

        hosts = [h for h in self._db_driver.hosts.values() if h.status == STATUS_ACTIVE]
        self.updater.add_hosts(hosts)
        self.updater.save()

        logger.info("Done")

    def clear(self):
        """Clear mrack records in /etc/hosts."""
        logger.info("Clearing mrack records in /etc/hosts.")

        self.updater.clear()
        self.updater.save()

        logger.info("Done")



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Beaker transformer module."""
import os
import re

from mrack.providers.provider import STRATEGY_ABORT
from mrack.transformers.transformer import Transformer

CONFIG_KEY = "beaker"
DEFAULT_ATTEMPTS = 2


class BeakerTransformer(Transformer):
    """Beaker transformer."""

    _config_key = CONFIG_KEY
    _required_config_attrs = [
        "distros",
        "reserve_duration",
        "timeout",
    ]  # List[str]

    async def init_provider(self):
        """Initialize associate provider and transformer display name."""
        self.dsp_name = "Beaker"
        await self._provider.init(
            distros=self.config["distros"].values(),
            timeout=self.config["timeout"],
            reserve_duration=self.config["reserve_duration"],
            strategy=self.config.get("strategy", STRATEGY_ABORT),
            max_retry=self.config.get("max_retry", DEFAULT_ATTEMPTS),
        )

    def _get_distro_and_variant(self, host):
        """Get distribution and its variant for the host system to requirement."""
        required_distro = self._find_value(
            host, "distro", "distros", host["os"], default=host["os"]
        )
        distro_variants = self.config.get("distro_variants")

        if "beaker_variant" in host:
            variant = host["beaker_variant"]
        elif distro_variants:
            variant = distro_variants.get(
                required_distro, distro_variants.get("default")
            )
        # keep this elif for backward compatibility wit mrack <= 1.2.0
        elif re.match(r"(rhel-[8|9])", host["os"]):
            variant = "BaseOS"
        else:  # Default to Server for RHEL7 and Fedora systems
            variant = "Server"

        return (required_distro, variant)

    def _get_ks_meta(self, host):
        """
        Get `ks_meta` value from host or provisioning config or default if not defined.

        The priority is following:
            - host
            - provisioning-config.yaml
            - default from provisioning config
            - empty if not defined in provisioning config

        """
        res = self._find_value(
            host.get(CONFIG_KEY, {}),
            "ks_meta",
            "kickstart_metadata",
            host["os"],
        )
        return res

    def _construct_ks_append_script(self, ks_append, pubkeys=None):
        """Create ks_appdend from requirements."""
        if not ks_append and not pubkeys:
            return []

        res_ks_append = ["%post"]
        if pubkeys:
            ks_append += self._allow_ssh_keys(pubkeys)

        res_ks_append += ks_append

        res_ks_append.append("%end")
        return ["\n".join(res_ks_append)]

    def _allow_ssh_keys(self, pubkeys):
        """Create ssh key content to be injected to xml."""
        keys_content = []
        keys_content.append("mkdir -p /root/.ssh")
        keys_content.append('cat >>/root/.ssh/authorized_keys << "__EOF__"')
        keys_content.append("# keys added by mrack:")

        for key in set(pubkeys):
            with open(os.path.expanduser(key), "r", encoding="utf-8") as key_file:
                keys_content.append(f"{key_file.read().strip()}")

        keys_content.append("# end section of keys added by mrack")
        keys_content.append("__EOF__")
        keys_content.append("restorecon -R /root/.ssh")
        keys_content.append("chmod go-w /root /root/.ssh /root/.ssh/authorized_keys")

        return keys_content

    def _get_pubkeys(self, host):
        """Get public keys list defined per host."""
        host_beaker_config = host.get(CONFIG_KEY, {})
        pubkey = self._find_value(host_beaker_config, "pubkey", "pubkey", None, None)

        # add support for more keys defined in list
        pubkeys = self._find_value(
            host_beaker_config, "pubkeys", "pubkeys", host["os"], None
        )

        # normalize and merge them
        if not pubkeys:
            pubkeys = []
        if pubkey:
            pubkeys.append(pubkey)

        return pubkeys

    def create_host_requirement(self, host):
        """Create single input for Beaker provisioner."""
        distro, variant = self._get_distro_and_variant(host)
        specs = {
            "name": host["name"],
            "distro": distro,
            "os": host["os"],
            "group": host["group"],
            "meta_distro": "distro" in host,
            "arch": host.get("arch", "x86_64"),
            "variant": variant,
            "ks_meta": self._get_ks_meta(host),
            "retention_tag": self._find_value(
                host.get(CONFIG_KEY, {}),
                "retention_tag",
                "retention_tag",
                host["os"],
                default="audit",
            ),
            "product": self._find_value(
                host.get(CONFIG_KEY, {}),
                "product",
                "product",
                host["os"],
                default="[internal]",
            ),
            "whiteboard": self._find_value(
                host.get(CONFIG_KEY, {}),
                "whiteboard",
                "whiteboard",
                host["os"],
                default="This job has been created using mrack.",
            ),
            "priority": self._find_value(
                host.get(CONFIG_KEY, {}),
                "priority",
                "priority",
                host["os"],
                default="Normal",
            ),
            # Recipe task definition
            "tasks": self._find_value(
                host.get(CONFIG_KEY, {}),
                "tasks",
                "tasks",
                host["os"],
                default=[  # we use dummy task because beaker require a task in recipe
                    {"name": "/distribution/dummy", "role": "STANDALONE"}
                ],
            ),
            "ks_append": self._construct_ks_append_script(
                self._find_value(
                    host.get(CONFIG_KEY, {}),
                    "ks_append",
                    "ks_append",
                    host["os"],
                    default=[],
                ),
                self._get_pubkeys(host),
            ),
            # TODO: should have similar logic as _get_flavor
            "hostRequires": self._find_value(
                host.get(CONFIG_KEY, {}),
                "hostRequires",
                "hostRequires",
                host["group"],
                default=None,
            ),
            "distro_tags": self._find_value(
                host.get(CONFIG_KEY, {}),
                "distro_tags",
                "distro_tags",
                distro,
                default=[],
            ),
        }

        return specs
